import pathlib

import panel as pn
import param  # type: ignore

from lumen.filters.base import Filter
from lumen.pipeline import Pipeline
from lumen.state import state as lm_state
from lumen.transforms.base import Transform

from .base import WizardItem
from .fast import FastComponent
from .gallery import Gallery, GalleryItem
from .state import state

ASSETS_DIR = pathlib.Path(__file__).parent / 'assets'


class PipelineEditor(FastComponent):

    filter_type = param.Selector(doc="Select a Filter to add to this Pipeline.")

    transform_type = param.Selector(doc="Select a Transform to add to this Pipeline.")

    filter_items = param.List()

    filters = param.ClassSelector(class_=(list, dict), precedence=-1)

    transform_items = param.List()

    transforms = param.List(precedence=-1)

    preview = param.Parameter(allow_refs=False)

    sizing_mode = param.String(default='stretch_width')

    source = param.Parameter()

    table = param.Parameter()

    spec = param.Dict(precedence=-1)

    resize = param.Integer(default=0)

    margin = param.Integer(default=10)

    _preview_display = param.String(default='none')

    _scripts = {
        'resize': ['window.dispatchEvent(new Event("resize"))']
    }

    _template = """
    <div style="display: flex; flex-direction: column; ">
      <span style="font-size: 2em"><b>{{ name }}</b></span>
      <span style="font-size: 1em"><b>Pipeline driven by {{ table }} table in the {{ source }} source</b></span>
    </div>
    <div style="display: flex; margin-top: 1em;">
      <div style="display: grid; flex: auto;">
        <label for="filter-select-${id}">
          <b>{{ param.filter_type.label }}</b>
        </label>
        <fast-select id="filter-select" style="min-width: 150px;" value="${filter_type}">
          {% for ftype in param.filter_type.objects %}
          <fast-option id="filter-option-{{ loop.index0 }}" value="{{ ftype }}">{{ ftype.title() }}</fast-option>
          {% endfor %}
        </fast-select>
        <fast-tooltip anchor="filter-select-${id}">{{ param.filter_type.doc }}</fast-tooltip>
      </div>
      <fast-button id="add-filter" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_filter}">
        <b style="font-size: 2em;">+</b>
      </fast-button>
    </div>
    <div id="filters" style="display: flex; flex-wrap: wrap;">
       {% for fitem in filter_items %}
       <div id="filter-item" style="margin: 0.5em; width: 335px; border: var(--accent-fill-rest) 1px solid; border-radius: 0.5em;">
         ${fitem}
         <fast-button id="filt-remove-button-{{ loop.index0 }}" appearance="accent" onclick="${_remove_filter}" style="float: right; z-index: 100;">
           <b style="font-size: 2em;">-</b>
         </fast-button>
       </div>
       {% endfor %}
    </div>
    <div style="display: flex;">
      <div style="display: grid; flex: auto;">
        <label for="transform-select-${id}">
          <b>{{ param.transform_type.label }}</b>
        </label>
        <fast-select id="transform-select" style="min-width: 150px;" value="${transform_type}">
          {% for ftype in param.transform_type.objects %}
          <fast-option id="transform-option-{{ loop.index0 }}" value="{{ ftype }}">{{ ftype.title() }}</fast-option>
          {% endfor %}
        </fast-select>
        <fast-tooltip anchor="transform-select-${id}">{{ param.transform_type.doc }}</fast-tooltip>
      </div>
      <fast-button id="add-transform" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_transform}">
        <b style="font-size: 2em;">+</b>
      </fast-button>
    </div>
    <div id="transforms" style="display: flex; flex-wrap: wrap;">
       {% for titem in transform_items %}
       <div id="transform-item" style="margin: 0.5em; width: 335px; border: var(--accent-fill-rest) 1px solid; border-radius: 0.5em;">
         ${titem}
         <fast-button id="transform-remove-button-{{ loop.index0 }}" appearance="accent" onclick="${_remove_transform}" style="float: right; z-index: 100;">
           <b style="font-size: 2em;">-</b>
         </fast-button>
       </div>
       {% endfor %}
    </div>
    <fast-button id="preview-button" onclick="${_preview}" style="margin-left: auto; margin-top: 1em; z-index: 100;">
      Preview
    </fast-button>
    <fast-divider></faster-divider>
    <div id="preview-area" style="display: ${_preview_display}; margin-top: 2em;">
      <div id="preview" style="flex: auto; margin-left: 1em; overflow-y: auto;">${preview}</div>
    </div>
    """

    thumbnail = ASSETS_DIR / 'source.png'

    _child_config = {
        'filter_items': 'model',
        'transform_items': 'model'
    }

    def __init__(self, **params):
        spec = params.pop('spec', {})
        params.update(**{
            k: v for k, v in spec.items() if k in self.param and k not in params
        })
        filters = params.pop('filters', [])
        transforms = params.pop('transforms', [])
        super().__init__(spec=spec, filters=[], transforms=[], **params)
        theme = 'midnight' #if getattr(pn.config, 'theme', 'default') == 'dark' else 'simple'
        self.preview = pn.widgets.Tabulator(
            sizing_mode='stretch_width', pagination='remote', page_size=12,
            theme=theme, height=400
        )
        self._pipeline = None
        self._populate_selectors()
        for filt in (filters.values() if isinstance(filters, dict) else filters):
            self._add_filter(spec=filt)
        for trans in transforms:
            self._add_transform(spec=trans)

    def _populate_selectors(self):
        filters = param.concrete_descendents(Filter).values()
        self.param.filter_type.objects = ftypes = [f.filter_type for f in filters if f.filter_type]
        self.filter_type = ftypes[0]
        transforms = param.concrete_descendents(Transform).values()
        self.param.transform_type.objects = ttypes = [
            t.transform_type for t in transforms if t.transform_type
        ]
        self.transform_type = ttypes[0]

    @param.depends('_preview_display', watch=True)
    def _update_displayed(self):
        if self._preview_display == 'none' or self._pipeline is not None:
            return
        self._pipeline = Pipeline.from_spec({
            k: v for k, v in self.spec.items() if k not in ('filters', 'transforms')
        })
        filters = self.filters.values() if isinstance(self.filters, dict) else self.filters
        for filt in filters or []:
            if not isinstance(filt, Filter):
                filt = Filter.from_spec(filt, {self._pipeline.table: self._pipeline.schema})
            self._pipeline.add_filter(filt)
        for trans in self.transforms or []:
            if not isinstance(trans, Filter):
                trans = Transform.from_spec(trans)
            self._pipeline.add_transform(trans)
        self._pipeline.param.watch(self._update_preview, ['data'])
        self.preview.value = self._pipeline.data

    def _update_preview(self, event):
        self.preview.value = event.new

    def _update_spec(self, *events):
        self.spec['filters'] = []
        for filt in self.filters:
            filt_spec = filt.to_spec()
            self.spec['filters'].append(filt_spec)
        self.spec['transforms'] = []
        for transform in self.transforms:
            transform_spec = transform.to_spec()
            self.spec['transforms'].append(transform_spec)

    def _preview(self, event):
        self._preview_display = 'flex' if self._preview_display == 'none' else 'none'
        self.resize += 1

    def _remove_filter(self, event):
        index = int(event.node.split('-')[-1])
        self.filters.pop(index)
        self.filter_items.pop(index)
        self.param.trigger('filters')
        self.param.trigger('filter_items')
        self._update_spec()

    def _remove_transform(self, event):
        index = int(event.node.split('-')[-1])
        self.transforms.pop(index)
        self.transform_items.pop(index)
        self.param.trigger('transforms')
        self.param.trigger('transform_items')
        self._update_spec()

    def _add_filter(self, event=None, spec=None):
        self.loading = True
        source = lm_state.sources[self.source]
        schema = source.get_schema(self.table)
        if spec is None:
            spec = {
                'type': self.filter_type,
                'table': self.table
            }
            if schema:
                spec['field'] = next(iter(schema))
        filt = Filter.from_spec(spec, {self.table: schema})
        filt.param.watch(self._update_spec, list(filt.param))
        if self._pipeline:
            self._pipeline.add_filter(filt)

        self.filters.append(filt)
        self.param.trigger('filters')
        params = [p for p in filt.param if p not in ('name', 'schema', 'table', 'widget')]
        self.filter_items.append(pn.Param(filt, parameters=params))
        self.param.trigger('filter_items')
        self._update_spec()
        self.loading = False

    def _add_transform(self, event=None, spec=None):
        self.loading = True
        source = lm_state.sources[self.source]
        schema = source.get_schema(self.table)
        if spec is None:
            spec = {'type': self.transform_type}

        transform = Transform.from_spec(spec)
        transform.param.watch(self._update_spec, list(transform.param))

        for p in transform._field_params:
            transform.param[p].objects = list(schema)

        self.transforms.append(transform)
        if self._pipeline:
            self._pipeline.add_transform(transform)
        self.param.trigger('transforms')
        self.transform_items.append(pn.panel(transform))
        self.param.trigger('transform_items')
        self._update_spec()
        self.loading = False

    def _save(self):
        pass


class PipelinesEditor(WizardItem):
    """
    Declare the data pipelines for your dashboard by giving your pipeline
    a name and adding the source.
    """

    disabled = param.Boolean(default=True)

    pipeline_name = param.String(doc="Enter a name for the pipeline")

    pipelines = param.Dict(default={}, doc="The list of pipelines added to the dashboard.")

    sources = param.List(label='Select a source', doc="""
        Select from list of sources (if None are defined ensure you select at least one Source
        in the source gallery.""")

    source = param.String()

    tables = param.List(label='Select a table', doc="Selects the table driving the pipeline.")

    table = param.String()

    _template = """
    <span style="font-size: 2em">Pipeline Editor</span>
    <p>{{ __doc__ }}</p>
    <fast-divider></fast-divider>
    <div style="display: flex;">
      <form role="form" style="flex: 30%; max-width: 450px; line-height: 2em;">
        <div style="display: grid;">
          <label for="pipeline-name-${id}"><b>{{ param.pipeline_name.label }}</b></label>
          <fast-text-field id="pipeline-name" placeholder="{{ param.pipeline_name.doc }}" value="${pipeline_name}">
          </fast-text-field>
        </div>
        <div style="display: grid;">
          <label for="sources-${id}"><b>{{ param.sources.label }}</b></label>
          <fast-select id="sources" style="max-width: 450px; min-width: 150px;" value="${source}">
          {% for src in sources %}
            <fast-option value="{{ src }}">{{ src.title() }}</fast-option>
          {% endfor %}
          </fast-select>
          <fast-tooltip anchor="sources-${id}">{{ param.sources.doc }}</fast-tooltip>
        </div>
        <div style="display: flex;">
          <div style="display: grid;">
            <label for="tables-${id}"><b>{{ param.tables.label }}</b></label>
            <fast-select id="tables" style="max-width: 450px; min-width: 150px;" value="${table}">
            {% for tbl in tables %}
              <fast-option value="{{ tbl|tojson }}">{{ tbl }}</fast-option>
            {% endfor %}
            </fast-select>
            <fast-tooltip anchor="tables-${id}">{{ param.tables.doc }}</fast-tooltip>
          </div>
          <fast-button id="submit" appearance="accent" style="margin-top: auto; margin-left: 1em; width: 20px;" onclick="${_add_pipeline}" disabled="${disabled}">
            <b style="font-size: 2em;">+</b>
          </fast-button>
        </div>
      </form>
      <div id="pipelines" style="flex: 70%; margin-left: 1em; margin-right: 2em;">
        {% for pipeline in pipelines.values() %}
        <div id="pipeline-container" style="overflow: auto;">${pipeline}</div>
        <fast-divider></faster-divider>
        {% endfor %}
      </div>
    </div>
    """

    _dom_events = {'pipeline-name': ['keyup']}

    def __init__(self, **params):
        super().__init__(**params)
        state.sources.param.watch(self._update_sources, 'items')
        self._watchers = {}
        self._update_sources()

    @param.depends('source', watch=True, on_init=True)
    def _update_tables(self):
        if self.source in lm_state.sources:
            self.tables = lm_state.sources[self.source].get_tables()
            if not self.table and self.tables:
                self.table = self.tables[0]

    def _update_sources(self, *events):
        for name, item in state.sources.items.items():
            if name not in self._watchers:
                self._watchers[name] = item.param.watch(self._update_sources, 'selected')
        self.sources = [source for source, item in state.sources.items.items() if item.selected]
        if not self.source and self.sources:
            self.source = self.sources[0]

    @param.depends('pipeline_name', 'table', watch=True)
    def _enable_add(self):
        self.disabled = not bool(self.pipeline_name and self.table)

    def _add_pipeline(self, event):
        self.spec[self.pipeline_name] = spec = {
            'name': self.pipeline_name,
            'source': self.source,
            'table': self.table
        }
        editor = PipelineEditor(
            name=self.pipeline_name, spec=spec, source=self.source,
            sizing_mode='stretch_width', table=self.table
        )
        self.pipelines[self.pipeline_name] = editor
        self.param.trigger('pipelines')
        self.pipeline_name = ''
        self.ready = True


class PipelineGalleryItem(GalleryItem):

    disabled = param.Boolean(default=False)

    editor = param.ClassSelector(class_=PipelineEditor, precedence=-1)

    thumbnail = param.Filename()

    def __init__(self, **params):
        if 'description' not in params:
            spec = params['editor'].spec
            params['description'] = f"Pipeline driven by {spec['table']} on {spec['source']} source."
        super().__init__(**params)
        self.view = pn.pane.PNG(self.thumbnail, height=200, max_width=300, align='center')
        self._modal_content = [self.editor]

    @param.depends('selected', watch=True, on_init=True)
    def _add_spec(self):
        pipelines = state.spec['pipelines']
        if self.selected:
            pipelines[self.name] = self.spec
            lm_state.pipelines[self.name] = Pipeline.from_spec(dict(self.spec, name=self.name))
        elif self.name in pipelines:
            del pipelines[self.name]
            del lm_state.pipelines[self.name]


class PipelineGallery(WizardItem, Gallery):
    """
    Select the pipelines to add to your dashboard application.
    """

    path = param.Foldername()

    pipelines = param.Dict(default={}, doc="The list of pipelines added to the dashboard.", precedence=-1)

    _template = """
    <span style="font-size: 1.5em">Pipelines</span>
    <fast-divider></fast-divider>
    <span style="font-size: 1.2em; font-weight: bold;">{{ __doc__ }}</p>
    <div id="items" style="margin: 1em 0; display: flex; flex-wrap: wrap; gap: 1em;">
      {% for item in items.values() %}
      <fast-card id="pipeline-container" style="width: 350px; height: 400px; margin: 0.2em;">
        ${item}
      </fast-card>
      {% endfor %}
      <fast-card id="pipelines-container-new" style="height: 400px; width: 350px; padding: 1em; margin: 0.2em;">
        <div style="display: grid;">
          <span style="font-size: 1.25em; font-weight: bold;">Add new pipeline</span>
          <i id="add-button" onclick="${_open_modal}" class="fa fa-plus" style="font-size: 14em; margin: 0.2em auto;" aria-hidden="true"></i>
        </div>
      </fast-card>
    </div>
    """

    _gallery_item = PipelineGalleryItem

    _editor_type = PipelineEditor

    def __init__(self, **params):
        super().__init__(**params)
        self._items = self.items
        self._editor = PipelinesEditor(spec={}, margin=10)
        self._save_button = pn.widgets.Button(name='Save pipelines')
        self._save_button.on_click(self._save_pipelines)
        self._modal_content = [self._editor, self._save_button]
        self._watchers = {}
        state.sources.param.watch(self._update_items, 'items')
        self._update_items()
        for name, item in self._items.items():
            self.pipelines[name] = item.editor

    @param.depends('spec', watch=True)
    def _update_params(self):
        for name, pipeline in self.spec.items():
            self.pipelines[name] = editor = PipelineEditor(name=name, spec=pipeline)
            self.items[name] = PipelineGalleryItem(
                name=name, spec=pipeline, selected=True, editor=editor,
                thumbnail=editor.thumbnail
            )
        self.param.trigger('pipelines')
        self.param.trigger('items')

    def _update_items(self, *events):
        for name, item in state.sources.items.items():
            if name not in self._watchers:
                self._watchers[name] = item.param.watch(self._update_items, 'selected')
        for name, item in self.items.items():
            if name not in self._items:
                self._items[name] = item
        self.items = {
            name: item for name, item in self._items.items()
            if (item.spec.get('source') in lm_state.sources or
                item.spec.get('pipeline') in lm_state.pipelines)
        }
        self.param.trigger('items')

    def _add_pipeline(self, event):
        state.modal[:] = self._modal_content
        self._editor._update_tables()
        state.template.open_modal()

    def _save_pipelines(self, event):
        for name, pipeline in self._editor.pipelines.items():
            self.spec[name] = pipeline.spec
            item = PipelineGalleryItem(
                name=name, spec=pipeline.spec, margin=0, selected=True,
                editor=pipeline, thumbnail=pipeline.thumbnail
            )
            self._items[name] = item
            self.items[name] = item
            self.pipelines[name] = pipeline
        self.param.trigger('items')
        self.param.trigger('pipelines')
        self._editor.pipelines = {}
        state.template.close_modal()



from __future__ import annotations

import asyncio
import json
import os

from functools import partial
from types import SimpleNamespace
from typing import Any, Literal, TypedDict

import instructor
import panel as pn
import param

from instructor.dsl.partial import Partial
from instructor.patch import Mode, patch
from pydantic import BaseModel

from .components import StatusBadge
from .interceptor import Interceptor
from .models import YesNo
from .utils import format_exception, log_debug, truncate_string


class Message(TypedDict):
    role: Literal["system", "user", "assistant"]
    content: str
    name: str | None

BASE_MODES = list(Mode)


class Llm(param.Parameterized):
    """
    Baseclass for LLM implementations.

    An LLM implementation wraps a local or cloud based LLM provider
    with instructor to enable support for correctly validating Pydantic
    models.
    """

    create_kwargs = param.Dict(default={}, doc="""
        Additional keyword arguments to pass to the LLM provider
        when calling chat.completions.create.""")

    mode = param.Selector(default=Mode.JSON_SCHEMA, objects=BASE_MODES, doc="""
        The calling mode used by instructor to guide the LLM towards generating
        outputs matching the schema.""")

    interceptor = param.ClassSelector(default=None, class_=Interceptor, doc="""
        Intercepter instance to intercept LLM calls, e.g. for logging.""")

    model_kwargs = param.Dict(default={}, doc="""
        LLM model definitions indexed by type. Supported types include
        'default', 'reasoning' and 'sql'. Agents may pick which model to
        invoke for different reasons.""")

    # Whether the LLM supports streaming of any kind
    _supports_stream = True

    # Whether the LLM supports streaming of Pydantic model output
    _supports_model_stream = True

    __abstract = True

    def __init__(self, **params):
        if "mode" in params:
            if isinstance(params["mode"], str):
                params["mode"] = Mode[params["mode"].upper()]
        super().__init__(**params)
        self._status = StatusBadge(name="LLM Pending")
        if not self.model_kwargs.get("default"):
            raise ValueError(
                f"Please specify a 'default' model in the model_kwargs "
                f"parameter for {self.__class__.__name__}."
            )

    def _get_model_kwargs(self, model_spec: str | dict) -> dict[str, Any]:
        """
        Can specify model kwargs as a dict or as a string that is a key in the model_kwargs
        or as a string that is a model type; else the actual name of the model.
        """
        if isinstance(model_spec, dict):
            return model_spec

        model_kwargs = self.model_kwargs.get(model_spec) or self.model_kwargs["default"]
        log_debug(f"LLM Model: \033[96m{model_kwargs.get('model')!r}\033[0m")
        return dict(model_kwargs)

    @property
    def _client_kwargs(self) -> dict[str, Any]:
        return {}

    def _add_system_message(
        self,
        messages: list[Message],
        system: str,
        input_kwargs: dict[str, Any]
    ) -> tuple[list[Message], dict[str, Any]]:
        if system:
            messages = [{"role": "system", "content": system}] + messages
        return messages, input_kwargs

    @classmethod
    def warmup(cls, model_kwargs: dict | None):
        """
        Allows LLM provider to perform actions that ensure that
        the model(s) are ready to run, e.g. downloading the model
        files.
        """

    async def invoke(
        self,
        messages: list[Message],
        system: str = "",
        response_model: BaseModel | None = None,
        allow_partial: bool = False,
        model_spec: str | dict = "default",
        **input_kwargs,
    ) -> BaseModel:
        """
        Invokes the LLM and returns its response.

        Arguments
        ---------
        messages: list[Message]
            A list of messages to feed to the LLM.
        system: str
            A system message to provide to the LLM.
        response_model: BaseModel | None
            A Pydantic model that the LLM should materialize.
        allow_partial: bool
            Whether to allow the LLM to only partially fill
            the provided response_model.
        model: Literal['default' | 'reasoning' | 'sql']
            The model as listed in the model_kwargs parameter
            to invoke to answer the query.

        Returns
        -------
        The completed response_model.
        """
        system = system.strip().replace("\n\n", "\n")
        messages, input_kwargs = self._add_system_message(messages, system, input_kwargs)

        kwargs = dict(self._client_kwargs)
        kwargs.update(input_kwargs)

        if response_model is not None:
            if allow_partial:
                response_model = Partial[response_model]
            kwargs["response_model"] = response_model

        output = await self.run_client(model_spec, messages, **kwargs)
        if output is None or output == "":
            raise ValueError("LLM failed to return valid output.")
        return output

    @classmethod
    def _get_delta(cls, chunk) -> str:
        if chunk.choices:
            return chunk.choices[0].delta.content or ""
        return ""

    def status(self):
        return self._status

    async def initialize(self, log_level: str):
        try:
            self._status.status = "running"
            await self.invoke(
                messages=[{'role': 'user', 'content': 'Are you there? YES | NO'}],
                model_spec="ui",
                response_model=YesNo
            )
            self._status.param.update(status="success", name='LLM Ready')
        except Exception as e:
            self._status.param.update(
                status="failed",
                name="LLM Not Connected",
                description='❌ '+(format_exception(e, limit=3) if log_level == 'DEBUG' else "Failed to connect to LLM"),
            )
            raise e

    async def stream(
        self,
        messages: list[Message],
        system: str = "",
        response_model: BaseModel | None = None,
        field: str | None = None,
        model_spec: str | dict = "default",
        **kwargs,
    ):
        """
        Invokes the LLM and streams its response.

        Arguments
        ---------
        messages: list[Message]
            A list of messages to feed to the LLM.
        system: str
            A system message to provide to the LLM.
        response_model: BaseModel | None
            A Pydantic model that the LLM should materialize.
        field: str
            The field in the response_model to stream.
        model: Literal['default' | 'reasoning' | 'sql']
            The model as listed in the model_kwargs parameter
            to invoke to answer the query.

        Yields
        ------
        The string or response_model field.
        """
        if ((response_model and not self._supports_model_stream) or
            not self._supports_stream):
            yield await self.invoke(
                messages,
                system=system,
                response_model=response_model,
                model_spec=model_spec,
                **kwargs,
            )
            return

        string = ""
        chunks = await self.invoke(
            messages,
            system=system,
            response_model=response_model,
            stream=True,
            allow_partial=True,
            model_spec=model_spec,
            **kwargs,
        )
        try:
            async for chunk in chunks:
                if response_model is None:
                    delta = self._get_delta(chunk)
                    string += delta
                    yield string
                else:
                    yield getattr(chunk, field) if field is not None else chunk
        except TypeError:
            for chunk in chunks:
                if response_model is None:
                    delta = self._get_delta(chunk)
                    string += delta
                    yield string
                else:
                    yield getattr(chunk, field) if field is not None else chunk

    async def run_client(self, model_spec: str | dict, messages: list[Message], **kwargs):
        log_debug(f"Input messages: \033[95m{len(messages)} messages\033[0m including system")
        previous_role = None
        for i, message in enumerate(messages):
            role = message["role"]
            if role == "system":
                continue
            if role == "user":
                log_debug(f"Message \033[95m{i} (u)\033[0m: {message['content']}")
            else:
                log_debug(f"Message \033[95m{i} (a)\033[0m: {message['content']}")
            if previous_role == role:
                log_debug(
                    "\033[91mWARNING: Two consecutive messages from the same role; "
                    "some providers disallow this.\033[0m"
                )
            previous_role = role

        client = await self.get_client(model_spec, **kwargs)
        result = await client(messages=messages, **kwargs)
        if response_model := kwargs.get("response_model"):
            log_debug(f"Response model: \033[93m{response_model.__name__!r}\033[0m")
        log_debug(f"LLM Response: \033[95m{truncate_string(str(result), max_length=1000)}\033[0m\n---")
        return result


class LlamaCpp(Llm):
    """
    A LLM implementation using Llama.cpp Python wrapper together
    with huggingface_hub to fetch the models.
    """

    chat_format = param.String(constant=True)

    temperature = param.Number(default=0.4, bounds=(0, None), constant=True)

    mode = param.Selector(default=Mode.JSON_SCHEMA, objects=BASE_MODES)

    model_kwargs = param.Dict(default={
        "default": {
            "repo": "Qwen/Qwen2.5-Coder-7B-Instruct-GGUF",
            "model_file": "qwen2.5-coder-7b-instruct-q5_k_m.gguf",
            "chat_format": "qwen",
        },
    })

    def _get_model_kwargs(self, model_spec: str | dict) -> dict[str, Any]:
        if isinstance(model_spec, dict):
            return model_spec

        model_kwargs = self.model_kwargs["default"]
        if model_spec in self.model_kwargs or "/" not in model_spec:
            model_kwargs = super()._get_model_kwargs(model_kwargs)
        else:
            repo, model_spec = model_spec.rsplit("/", 1)
            if ":" in model_spec:
                model_file, chat_format = model_spec.split(":")
                model_kwargs["chat_format"] = chat_format
            else:
                model_file = model_spec
            model_kwargs["repo"] = repo
            model_kwargs["model_file"] = model_file

        if "n_ctx" not in model_kwargs:
            # 0 = from model
            model_kwargs["n_ctx"] = 0
        return dict(model_kwargs)

    @property
    def _client_kwargs(self) -> dict[str, Any]:
        return {"temperature": self.temperature}

    def _cache_model(self, model_spec: str | dict, mode: str, **kwargs):
        from llama_cpp import Llama as LlamaCpp
        llm = LlamaCpp(**kwargs)

        raw_client = llm.create_chat_completion_openai_v1
        # patch works with/without response_model
        client_callable = patch(create=raw_client, mode=mode)
        pn.state.cache[(model_spec, mode)] = client_callable
        return client_callable

    @classmethod
    def warmup(cls, model_kwargs: dict | None):
        model_kwargs = model_kwargs or cls.model_kwargs
        if 'default' not in model_kwargs:
            model_kwargs['default'] = cls.model_kwargs['default']
        huggingface_models = {
            model: llm_spec for model, llm_spec in model_kwargs.items()
            if 'repo' in llm_spec or 'repo_id' in llm_spec
        }
        if not huggingface_models:
            return

        from huggingface_hub import hf_hub_download
        print(f"{cls.__name__} provider is downloading following models:\n\n{json.dumps(huggingface_models, indent=2)}")
        for model, kwargs in model_kwargs.items():
            repo = kwargs.get('repo', kwargs.get('repo_id'))
            model_file = kwargs.get('model_file')
            hf_hub_download(repo, model_file)

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        model_kwargs = self._get_model_kwargs(model_spec)
        mode = model_kwargs.pop("mode", self.mode)
        if client_callable := pn.state.cache.get((model_spec, mode)):
            return client_callable
        if 'repo' in model_kwargs:
            from huggingface_hub import hf_hub_download
            repo = model_kwargs.get('repo', model_kwargs.get('repo_id'))
            model_file = model_kwargs.get('model_file')
            model_path = await asyncio.to_thread(hf_hub_download, repo, model_file)
        elif 'model_path' in model_kwargs:
            model_path = model_kwargs['model_path']
        else:
            raise ValueError(
                "LlamaCpp.model_kwargs must contain either a 'repo' and 'model_file' "
                "(to fetch a model using `huggingface_hub` or a 'model_path' pointing "
                "to a model on disk."
            )
        llm_kwargs = dict(
            model_path=model_path,
            n_gpu_layers=-1,
            seed=128,
            logits_all=False,
            use_mlock=True,
            verbose=False,
        )
        llm_kwargs.update(model_kwargs)
        client_callable = await asyncio.to_thread(self._cache_model, model_spec, mode=mode, **llm_kwargs)
        return client_callable

    async def run_client(self, model_spec: str | dict, messages: list[Message], **kwargs):
        client = await self.get_client(model_spec, **kwargs)
        return client(messages=messages, **kwargs)


class OpenAI(Llm):
    """
    An LLM implementation using the OpenAI cloud.
    """

    api_key = param.String(doc="The OpenAI API key.")

    endpoint = param.String(doc="The OpenAI API endpoint.")

    mode = param.Selector(default=Mode.TOOLS)

    temperature = param.Number(default=0.25, bounds=(0, None), constant=True)

    organization = param.String(doc="The OpenAI organization to charge.")

    model_kwargs = param.Dict(default={
        "default": {"model": "gpt-4o-mini"},
        "sql": {"model": "gpt-4.1-mini"},
        "reasoning": {"model": "gpt-4o"},
    })

    use_logfire = param.Boolean(default=False, doc="""
        Whether to log LLM calls and responses to logfire.""")

    @property
    def _client_kwargs(self):
        return {"temperature": self.temperature}

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        import openai

        model_kwargs = self._get_model_kwargs(model_spec)
        model = model_kwargs.pop("model")
        mode = model_kwargs.pop("mode", self.mode)
        if self.endpoint:
            model_kwargs["base_url"] = self.endpoint
        if self.api_key:
            model_kwargs["api_key"] = self.api_key
        if self.organization:
            model_kwargs["organization"] = self.organization
        llm = openai.AsyncOpenAI(**model_kwargs)

        if self.interceptor:
            self.interceptor.patch_client(llm, mode="store_inputs")

        if response_model:
            llm = patch(llm, mode=mode)

        if self.interceptor:
            # must be called after instructor
            self.interceptor.patch_client_response(llm)

        client_callable = partial(llm.chat.completions.create, model=model, **self.create_kwargs)

        if self.use_logfire:
            import logfire
            logfire.configure()
            logfire.instrument_openai(llm)
        return client_callable


class AzureOpenAI(Llm):
    """
    A LLM implementation that uses the Azure OpenAI integration.
    """

    api_key = param.String(default=os.getenv("AZUREAI_ENDPOINT_KEY"), doc="The Azure API key.")

    api_version = param.String(default="2024-10-21", doc="The Azure AI Studio API version.")

    endpoint = param.String(default=os.getenv('AZUREAI_ENDPOINT_URL'), doc="The Azure AI Studio endpoint.")

    mode = param.Selector(default=Mode.TOOLS)

    model_kwargs = param.Dict(default={
        "default": {"model": "gpt-4o-mini"},
        "reasoning": {"model": "gpt-4o"},
    })

    temperature = param.Number(default=1, bounds=(0, None), constant=True)

    @property
    def _client_kwargs(self):
        return {"temperature": self.temperature}

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        import openai

        model_kwargs = self._get_model_kwargs(model_spec)
        model = model_kwargs.pop("model")
        mode = model_kwargs.pop("mode", self.mode)
        if self.api_version:
            model_kwargs["api_version"] = self.api_version
        if self.api_key:
            model_kwargs["api_key"] = self.api_key
        if self.endpoint:
            model_kwargs["azure_endpoint"] = self.endpoint
        llm = openai.AsyncAzureOpenAI(**model_kwargs)

        if self.interceptor:
            self.interceptor.patch_client(llm, mode="store_inputs")

        if response_model:
            llm = patch(llm, mode=mode)

        if self.interceptor:
            # must be called after instructor
            self.interceptor.patch_client_response(llm)

        client_callable = partial(llm.chat.completions.create, model=model, **self.create_kwargs)
        return client_callable


class MistralAI(Llm):
    """
    A LLM implementation that calls Mistral AI.
    """

    api_key = param.String(default=os.getenv("MISTRAL_API_KEY"), doc="The Mistral AI API key.")

    mode = param.Selector(default=Mode.MISTRAL_TOOLS, objects=[Mode.JSON_SCHEMA, Mode.MISTRAL_TOOLS])

    temperature = param.Number(default=0.7, bounds=(0, 1), constant=True)

    model_kwargs = param.Dict(default={
        "default": {"model": "mistral-small-latest"},
        "reasoning": {"model": "mistral-large-latest"},
    })

    _supports_model_stream = False  # instructor doesn't work with Mistral's streaming

    @property
    def _client_kwargs(self):
        return {"temperature": self.temperature}

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        from mistralai import Mistral

        model_kwargs = self._get_model_kwargs(model_spec)
        model_kwargs["api_key"] = self.api_key
        model = model_kwargs.pop("model")
        mode = model_kwargs.pop("mode", self.mode)
        llm = Mistral(**model_kwargs)

        stream = kwargs.get("stream", False)
        llm.chat.completions = SimpleNamespace(create=None)  # make it like OpenAI for simplicity
        llm.chat.completions.create = llm.chat.stream_async if stream else llm.chat.complete_async

        if self.interceptor:
            self.interceptor.patch_client(llm, mode="store_inputs")

        if response_model:
            llm = patch(llm, mode=mode)

        if self.interceptor:
            self.interceptor.patch_client_response(llm)

        client_callable = partial(llm.chat.completions.create, model=model, **self.create_kwargs)
        return client_callable

    @classmethod
    def _get_delta(cls, chunk):
        if chunk.data.choices:
            return chunk.data.choices[0].delta.content or ""
        return ""


class AzureMistralAI(MistralAI):
    """
    A LLM implementation that calls Mistral AI models on Azure.
    """

    api_key = param.String(default=os.getenv("AZUREAI_ENDPOINT_KEY"), doc="The Azure API key")

    endpoint = param.String(default=os.getenv('AZUREAI_ENDPOINT_URL'), doc="The Azure API endpoint to invoke.")

    model_kwargs = param.Dict(default={
        "default": {"model": "azureai"},
    })

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        from mistralai_azure import MistralAzure

        async def llm_chat_non_stream_async(*args, **kwargs):
            response = await llm.chat.complete_async(*args, **kwargs)
            return response.choices[0].message.content

        model_kwargs = self._get_model_kwargs(model_spec)
        model_kwargs["api_key"] = self.api_key
        model_kwargs["azure_endpoint"] = self.endpoint
        model = model_kwargs.pop("model")
        mode = model_kwargs.pop("mode", self.mode)
        llm = MistralAzure(**model_kwargs)

        stream = kwargs.get("stream", False)
        llm.chat.completions = SimpleNamespace(create=None)  # make it like OpenAI for simplicity
        llm.chat.completions.create = llm.chat.stream_async if stream else llm.chat.complete_async

        if self.interceptor:
            self.interceptor.patch_client(llm, mode="store_inputs")

        if response_model:
            llm = patch(llm, mode=mode)

        if self.interceptor:
            self.interceptor.patch_client_response(llm)

        client_callable = partial(llm.chat.completions.create, model=model, **self.create_kwargs)
        return client_callable


class AnthropicAI(Llm):
    """
    A LLM implementation that calls Anthropic models such as Claude.
    """

    api_key = param.String(default=os.getenv("ANTHROPIC_API_KEY"), doc="The Anthropic API key.")

    mode = param.Selector(default=Mode.ANTHROPIC_TOOLS, objects=[Mode.ANTHROPIC_JSON, Mode.ANTHROPIC_TOOLS])

    temperature = param.Number(default=0.7, bounds=(0, 1), constant=True)

    model_kwargs = param.Dict(default={
        "default": {"model": "claude-3-5-haiku-latest"},
        "reasoning": {"model": "claude-3-5-sonnet-latest"},
    })

    _supports_model_stream = True

    @property
    def _client_kwargs(self):
        return {"temperature": self.temperature, "max_tokens": 1024}

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        if self.interceptor:
            raise NotImplementedError("Interceptors are not supported for AnthropicAI.")

        from anthropic import AsyncAnthropic

        model_kwargs = self._get_model_kwargs(model_spec)
        model = model_kwargs.pop("model")

        llm = AsyncAnthropic(api_key=self.api_key, **model_kwargs)

        if response_model:
            client = instructor.from_anthropic(llm)
            return partial(client.messages.create, model=model, **self.create_kwargs)
        else:
            return partial(llm.messages.create, model=model, **self.create_kwargs)

    @classmethod
    def _get_delta(cls, chunk: Any) -> str:
        if hasattr(chunk, 'delta'):
            if hasattr(chunk.delta, "text"):
                return chunk.delta.text
        return ""

    def _add_system_message(self, messages: list[Message], system: str, input_kwargs: dict[str, Any]):
        input_kwargs["system"] = system
        return messages, input_kwargs


class AINavigator(OpenAI):
    """
    A LLM implementation that calls the [Anaconda AI Navigator](https://www.anaconda.com/products/ai-navigator) API.
    """

    endpoint = param.String(
        default="http://localhost:8080/v1", doc="""
            The API endpoint; should include the full address, including the port.""")

    mode = param.Selector(default=Mode.JSON_SCHEMA)


class MessageModel(BaseModel):
    role: Literal["system", "user", "assistant"]
    content: str
    name: str | None


class Choice(BaseModel):

    delta: MessageModel | None

    message: MessageModel | None

    finish_reason: str | None


class Response(BaseModel):

    choices: list[Choice]


class WebLLM(Llm):

    model_kwargs = param.Dict({'default': {'model_slug': 'Qwen2.5-7B-Instruct-q4f16_1-MLC'}})

    mode = param.Parameter(default=Mode.JSON_SCHEMA)

    temperature = param.Number(default=0.4, bounds=(0, None))

    def __init__(self, **params):
        from panel_web_llm import WebLLM as pnWebLLM
        self._llm = pnWebLLM()
        super().__init__(**params)

    async def _create_completion(self, messages, **kwargs):
        if kwargs.get('stream', False):
            async def generator():
                async for chunk in self._llm.create_completion(messages, **kwargs):
                    yield Response(choices=[
                        Choice(
                            delta=MessageModel(
                                content=chunk['delta']['content'],
                                role=chunk['delta']['role'],
                                name=None
                            ) if chunk['delta'] else None,
                            message=None,
                            finish_reason=chunk['finish_reason']
                        )
                    ])
        else:
            content, reason, role = "", "", ""
            async for chunk in self._llm.create_completion(messages, **kwargs):
                content = chunk['message']['content']
                role = chunk['message']['role']
                reason = chunk['finish_reason']
            msg = MessageModel(content=content, name=None, role=role)
            return Response(choices=[Choice(message=msg, delta=None, finish_reason=reason)])
        return generator()

    async def get_client(self, model_spec: str | dict, response_model: BaseModel | None = None, **kwargs):
        model_kwargs = self._get_model_kwargs(model_spec)
        mode = model_kwargs.pop("mode", self.mode)
        if not self._llm.loaded:
            self._llm.param.update(**{k: v for k, v in model_kwargs.items() if k in self._llm.param})
            self._llm.param.trigger('load_model')
        client_callable = patch(create=self._create_completion, mode=mode)
        return client_callable

    def status(self):
        return pn.Row(self._status, self._llm)

    async def initialize(self, log_level: str):
        await asyncio.sleep(2)
        progress = self._llm.param.load_status.rx().get("progress", 0)*100
        self._status.name = pn.rx('Loading LLM {:.1f}%').format(progress)
        try:
            await self.invoke(
                messages=[{'role': 'user', 'content': 'Are you there? YES | NO'}],
                model_spec="ui",
                response_model=YesNo
            )
        except Exception as e:
            self._status.param.update(
                status="failed",
                name="LLM Not Connected",
                description='❌ '+(format_exception(e, limit=3) if log_level == 'DEBUG' else "Failed to connect to LLM"),
            )
            raise e
        else:
            self._status.param.update(status="success", name='LLM Ready')



from __future__ import annotations

import asyncio
import json
import re
import traceback

from copy import deepcopy
from functools import partial
from typing import Any, Literal

import pandas as pd
import panel as pn
import param
import yaml

from panel.chat import ChatInterface
from panel.layout import Column
from panel.viewable import Viewable, Viewer
from pydantic import BaseModel, create_model
from pydantic.fields import FieldInfo

from ..base import Component
from ..dashboard import Config
from ..pipeline import Pipeline
from ..sources.base import BaseSQLSource, Source
from ..sources.duckdb import DuckDBSource
from ..state import state
from ..transforms.sql import SQLLimit
from ..views import (
    Panel, VegaLiteView, View, hvPlotUIView,
)
from .actor import ContextProvider
from .config import (
    PROMPTS_DIR, SOURCE_TABLE_SEPARATOR, VEGA_MAP_LAYER,
    VEGA_ZOOMABLE_MAP_ITEMS, RetriesExceededError,
)
from .controls import RetryControls, SourceControls
from .llm import Llm, Message
from .memory import _Memory
from .models import (
    DbtslQueryParams, PartialBaseModel, RetrySpec, Sql, VegaLiteSpec,
    make_find_tables_model,
)
from .schemas import get_metaset
from .services import DbtslMixin
from .tools import ToolUser
from .translate import param_to_pydantic
from .utils import (
    clean_sql, describe_data, get_data, get_pipeline, get_schema, load_json,
    log_debug, mutate_user_message, parse_table_slug, report_error,
    retry_llm_output, stream_details,
)
from .views import (
    AnalysisOutput, LumenOutput, SQLOutput, VegaLiteOutput,
)


class Agent(Viewer, ToolUser, ContextProvider):
    """
    Agents are actors responsible for taking a user query and
    performing a particular task and responding by adding context to
    the current memory and creating outputs.

    Each Agent can require certain context that is needed to perform
    the task and should declare any context it itself provides.

    Agents have access to an LLM and the current memory and can
    solve tasks by executing a series of prompts or by rendering
    contents such as forms or widgets to gather user input.
    """

    debug = param.Boolean(default=False, doc="""
        Whether to enable verbose error reporting.""")

    llm = param.ClassSelector(class_=Llm, doc="""
        The LLM implementation to query.""")

    steps_layout = param.ClassSelector(default=None, class_=Column, allow_None=True, doc="""
        The layout progress updates will be streamed to.""")

    user = param.String(default="Agent", doc="""
        The name of the user that will be respond to the user query.""")

    # Panel extensions this agent requires to be loaded
    _extensions = ()

    # Maximum width of the output
    _max_width = 1200

    __abstract = True

    def __init__(self, **params):
        def _exception_handler(exception):
            messages = self.interface.serialize()
            if messages and str(exception) in messages[-1]["content"]:
                return

            import traceback
            traceback.print_exception(exception)
            self.interface.send(
                f"Error cannot be resolved:\n\n{exception}",
                user="System",
                respond=False
            )

        if "interface" not in params:
            params["interface"] = ChatInterface(callback=self._interface_callback)
        super().__init__(**params)
        if not self.debug:
            pn.config.exception_handler = _exception_handler

        if state.config is None:
            state.config = Config(raise_with_notifications=True)
        else:
            state.config.raise_with_notifications = True

    async def _interface_callback(self, contents: list | str, user: str, instance: ChatInterface):
        await self.respond(contents)
        self._retries_left = 1

    @property
    def _steps_layout(self):
        """
        Stream to a dummy column to be able to suppress the steps output.
        """
        return pn.Column() if not self.steps_layout else self.steps_layout

    def __panel__(self):
        return self.interface

    async def _stream(self, messages: list[Message], system_prompt: str) -> Any:
        message = None
        model_spec = self.prompts["main"].get("llm_spec", self.llm_spec_key)
        async for output_chunk in self.llm.stream(
            messages, system=system_prompt, model_spec=model_spec, field="output"
        ):
            message = self.interface.stream(
                output_chunk, replace=True, message=message, user=self.user, max_width=self._max_width
            )
        return message

    async def _gather_prompt_context(self, prompt_name: str, messages: list, **context):
        context = await super()._gather_prompt_context(prompt_name, messages, **context)
        if "tool_context" not in context:
            context["tool_context"] = await self._use_tools(prompt_name, messages)
        return context

    # Public API

    @classmethod
    async def applies(cls, memory: _Memory) -> bool:
        """
        Additional checks to determine if the agent should be used.
        """
        return True

    async def requirements(self, messages: list[Message]) -> list[str]:
        return self.requires

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        """
        Provides a response to the user query.

        The type of the response may be a simple string or an object.

        Arguments
        ---------
        messages: list[Message]
            The list of messages corresponding to the user query and any other
            system messages to be included.
        render_output: bool
            Whether to render the output to the chat interface.
        step_title: str | None
            If the Agent response is part of a longer query this describes
            the step currently being processed.
        """
        system_prompt = await self._render_prompt("main", messages)
        return await self._stream(messages, system_prompt)


class SourceAgent(Agent):
    """
    SourceAgent renders a form that allows a user to upload or
    provide a URI to one or more datasets.
    """

    conditions = param.List(default=[
        "Use when user wants to upload or connect to new data sources, otherwise do not use",
        "Do not use to see if there are any relevant data sources available"
    ])

    purpose = param.String(default="The SourceAgent allows a user to upload new datasets, tables, or documents.")

    requires = param.List(default=[], readonly=True)

    provides = param.List(default=["sources", "source", "document_sources"], readonly=True)

    on_init = param.Boolean(default=True)

    _extensions = ('filedropper',)

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        source_controls = SourceControls(
            multiple=True, replace_controls=True, select_existing=False, memory=self.memory
        )

        output = pn.Column(source_controls)
        if "source" not in self._memory:
            help_message = "No datasets or documents were found, **please upload at least one to continue**..."
        else:
            help_message = "**Please upload new dataset(s)/document(s) to continue**, or click cancel if this was unintended..."
        output.insert(0, help_message)
        self.interface.send(output, respond=False, user="SourceAgent")
        while not source_controls._add_button.clicks > 0:
            await asyncio.sleep(0.05)
            if source_controls._cancel_button.clicks > 0:
                self.interface.undo()
                self.interface.disabled = False
                return None
        return source_controls


class ChatAgent(Agent):
    """
    ChatAgent provides general information about available data
    and other topics  to the user.
    """

    conditions = param.List(default=[
        "Best for high-level information about data or general conversation",
        "Can be used to describe available tables",
        "Not useful for answering data specific questions",
        "Must be paired with TableLookup or DocumentLookup",
    ])

    purpose = param.String(default="""
        Engages in conversations about high-level data topics,
        offering suggestions or insights to get started.""")

    prompts = param.Dict(
        default={
            "main": {
                "template": PROMPTS_DIR / "ChatAgent" / "main.jinja2",
            },
        }
    )

    requires = param.List(default=[], readonly=True)

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        context = {"tool_context": await self._use_tools("main", messages)}
        system_prompt = await self._render_prompt("main", messages, **context)
        return await self._stream(messages, system_prompt)


class AnalystAgent(ChatAgent):

    conditions = param.List(default=[
        "Best for interpreting query results from data output",
    ])

    purpose = param.String(default="""
        Responsible for analyzing results and providing clear, concise, and actionable insights.
        Focuses on breaking down complex data findings into understandable points for
        high-level decision-making. Emphasizes detailed interpretation, trends, and
        relationships within the data, while avoiding general overviews or
        superficial descriptions.""")

    prompts = param.Dict(
        default={
            "main": {
                "template": PROMPTS_DIR / "AnalystAgent" / "main.jinja2",
                "tools": []
            },
        }
    )

    requires = param.List(default=["source", "pipeline"], readonly=True)

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        messages = await super().respond(messages, render_output, step_title)
        if len(self._memory["data"]) == 0 and self._memory.get("sql"):
            self._memory["sql"] = f"{self._memory['sql']}\n-- No data was returned from the query."
        return messages

class ListAgent(Agent):
    """
    Abstract base class for agents that display a list of items to the user.
    """

    purpose = param.String(default="""
        Renders a list of items to the user and lets the user pick one.""")

    requires = param.List(default=[], readonly=True)

    _extensions = ('tabulator',)

    _column_name = None

    _message_format = None

    __abstract = True

    def _get_items(self) -> list[str]:
        """Return the list of items to display"""

    def _use_item(self, event):
        """Handle when a user clicks on an item in the list"""
        if event.column != "show":
            return
        item = self._df.iloc[event.row, 0]

        if self._message_format is None:
            raise ValueError("Subclass must define _message_format")

        message = self._message_format.format(item=repr(item))
        self.interface.send(message)

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        items = self._get_items()
        header_filters = False
        if len(items) > 10:
            column_filter = {"type": "input", "placeholder": f"Filter by {self._column_name.lower()}..."}
            header_filters = {self._column_name: column_filter}

        self._df = pd.DataFrame({self._column_name: items})
        item_list = pn.widgets.Tabulator(
            self._df,
            buttons={'show': '<i class="fa fa-eye"></i>'},
            show_index=False,
            min_height=150,
            min_width=350,
            widths={self._column_name: '90%'},
            disabled=True,
            page_size=10,
            pagination="remote",
            header_filters=header_filters,
            sizing_mode="stretch_width",
        )

        item_list.on_click(self._use_item)
        self.interface.stream(item_list, user="Lumen")
        return item_list


class TableListAgent(ListAgent):
    """
    The TableListAgent lists all available tables and lets the user pick one.
    """

    conditions = param.List(default=[
        "For listing available data tables in source to the user, but not for planning",
        "Not for showing data table contents",
    ])

    not_with = param.List(default=["DbtslAgent", "SQLAgent"])

    purpose = param.String(default="""
        Displays a list of all available tables in memory.""")

    requires = param.List(default=["source"], readonly=True)

    _column_name = "Table"

    _message_format = "Show the table: {item}"

    @classmethod
    async def applies(cls, memory: _Memory) -> bool:
        source = memory.get("source")
        if not source:
            return True  # source not loaded yet; always apply
        return len(source.get_tables()) > 1

    def _get_items(self) -> list[str]:
        tables = []
        if "closest_tables" in self._memory:
            tables = self._memory["closest_tables"]

        for source in self._memory["sources"]:
            tables += source.get_tables()

        # remove duplicates, keeps order in which they were added
        return pd.unique(tables).tolist()


class DocumentListAgent(ListAgent):
    """
    The DocumentListAgent lists all available documents provided by the user.
    """

    purpose = param.String(default="""
        Displays a list of all available documents in memory.""")

    requires = param.List(default=["document_sources"], readonly=True)

    _column_name = "Documents"

    _message_format = "Tell me about: {item}"

    @classmethod
    async def applies(cls, memory: _Memory) -> bool:
        sources = memory.get("document_sources")
        if not sources:
            return False  # source not loaded yet; always apply
        return len(sources) > 1

    def _get_items(self) -> list[str]:
        # extract the filename, following this pattern `Filename: 'filename'``
        return [doc["metadata"].get("filename", "untitled") for doc in self._memory["document_sources"]]


class LumenBaseAgent(Agent):

    user = param.String(default="Lumen")

    prompts = param.Dict(
        default={
            "retry_output": {
                "response_model": RetrySpec,
                "template": PROMPTS_DIR / "LumenBaseAgent" / "retry_output.jinja2"
            },
        }
    )

    _output_type = LumenOutput

    _max_width = None

    def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        """
        Update the specification in memory.
        """

    def _render_lumen(
        self,
        component: Component,
        message: pn.chat.ChatMessage = None,
        messages: list | None = None,
        render_output: bool = False,
        title: str | None = None,
        **kwargs
    ):
        memory = self._memory

        async def retry_invoke(event: param.parameterized.Event):
            reason = event.new
            modified_messages = mutate_user_message(
                f"New feedback: {reason!r}.\n\nThese were the previous instructions to use as reference:",
                deepcopy(messages), wrap='\n"""\n', suffix=False
            )
            with self.param.update(memory=memory):
                # TODO: only input the inner spec to retry
                system = await self._render_prompt(
                    "retry_output", messages=modified_messages, spec=out.spec,
                    language=out.language,
                )
            retry_model = self._lookup_prompt_key("retry_output", "response_model")
            with out.param.update(loading=True):
                result = await self.llm.invoke(
                    modified_messages, system=system, response_model=retry_model, model_spec="reasoning",
                )
                if "```" in result:
                    spec = re.search(r'(?s)```(\w+)?\s*(.*?)```', result.corrected_spec, re.DOTALL)
                else:
                    spec = result.corrected_spec
                out.spec = spec

        retry_controls = RetryControls()
        retry_controls.param.watch(retry_invoke, "reason")
        out = self._output_type(
            component=component,
            footer=[retry_controls],
            render_output=render_output,
            title=title,
            **kwargs
        )
        out.param.watch(partial(self._update_spec, self._memory), 'spec')
        if 'outputs' in self._memory:
            # We have to create a new list to trigger an event
            # since inplace updates will not trigger updates
            # and won't allow diffing between old and new values
            self._memory['outputs'] = self._memory['outputs']+[out]
        message_kwargs = dict(value=out, user=self.user)
        self.interface.stream(
            message=message, **message_kwargs, replace=True, max_width=self._max_width
        )


class SQLAgent(LumenBaseAgent):

    conditions = param.List(
        default=[
            "Start with this agent if you are unsure what to use",
            "For existing tables, only use if additional calculations are needed",
            "When reusing tables, reference by name rather than regenerating queries",
            "Not useful if the user is using the same data for plotting",
        ]
    )

    exclusions = param.List(default=["dbtsl_metaset"])

    not_with = param.List(default=["DbtslAgent", "TableLookup", "TableListAgent"])

    purpose = param.String(default="""
        Handles the display of tables and the creation, modification, and execution
        of SQL queries to address user queries about the data. Executes queries in
        a single step, encompassing tasks such as table joins, filtering, aggregations,
        and calculations. If additional columns are required, SQLAgent can join the
        current table with other tables to fulfill the query requirements.""")

    prompts = param.Dict(
        default={
            "main": {
                "response_model": Sql,
                "template": PROMPTS_DIR / "SQLAgent" / "main.jinja2",
            },
            "find_tables": {
                "response_model": make_find_tables_model,
                "template": PROMPTS_DIR / "SQLAgent" / "find_tables.jinja2"
            },
        }
    )

    provides = param.List(default=["table", "sql", "pipeline", "data"], readonly=True)

    requires = param.List(default=["source", "sql_metaset"], readonly=True)

    _extensions = ('codeeditor', 'tabulator',)

    _output_type = SQLOutput

    @retry_llm_output()
    async def _create_valid_sql(
        self,
        messages: list[Message],
        dialect: str,
        comments: str,
        title: str,
        tables_to_source: dict[str, BaseSQLSource],
        errors=None
    ):
        errors_context = {}
        if errors:
            # get the head of the tables
            columns_context = ""
            vector_metadata_map = self._memory["sql_metaset"].vector_metaset.vector_metadata_map
            for table_slug, vector_metadata in vector_metadata_map.items():
                table_name = table_slug.split(SOURCE_TABLE_SEPARATOR)[-1]
                if table_name in tables_to_source:
                    columns = [col.name for col in vector_metadata.columns or []]
                    columns_context += f"\nSQL: {vector_metadata.base_sql}\nColumns: {', '.join(columns)}\n\n"
            last_output = self._memory["sql"]
            num_errors = len(errors)
            errors = ('\n'.join(f"{i+1}. {error}" for i, error in enumerate(errors))).strip()
            errors_context = {
                "errors": errors,
                "last_output": last_output,
                "num_errors": num_errors,
                "columns_context": columns_context,
            }

        join_required = len(tables_to_source) > 1
        comments = comments if join_required else ""  # comments are about joins
        system_prompt = await self._render_prompt(
            "main",
            messages,
            join_required=join_required,
            dialect=dialect,
            comments=comments,
            **errors_context
        )
        with self._add_step(title=title or "SQL query", steps_layout=self._steps_layout) as step:
            model_spec = self.prompts["main"].get("llm_spec", self.llm_spec_key)
            response = self.llm.stream(messages, system=system_prompt, model_spec=model_spec, response_model=self._get_model("main"))
            sql_query = None
            try:
                async for output in response:
                    step_message = output.chain_of_thought or ""
                    step.stream(step_message, replace=True)
                if output.query:
                    sql_query = clean_sql(output.query, dialect)
                if sql_query and output.expr_slug:
                    step.stream(f"\n\n`{output.expr_slug}`\n```sql\n{sql_query}\n```")
                self._memory["sql"] = sql_query
            except asyncio.CancelledError as e:
                step.failed_title = "Cancelled SQL query generation"
                raise e

        if step.failed_title and step.failed_title.startswith("Cancelled"):
            raise asyncio.CancelledError()
        elif not sql_query:
            raise ValueError("No SQL query was generated.")

        sources = set(tables_to_source.values())
        if len(sources) > 1:
            mirrors = {}
            for a_table, a_source in tables_to_source.items():
                if not any(a_table.rstrip(")").rstrip("'").rstrip('"').endswith(ext)
                           for ext in [".csv", ".parquet", ".parq", ".json", ".xlsx"]):
                    renamed_table = a_table.replace(".", "_")
                    sql_query = sql_query.replace(a_table, renamed_table)
                else:
                    renamed_table = a_table
                # Remove source prefixes from table names
                if SOURCE_TABLE_SEPARATOR in renamed_table:
                    _, renamed_table = renamed_table.split(SOURCE_TABLE_SEPARATOR, maxsplit=1)
                sql_query = sql_query.replace(a_table, renamed_table)
                mirrors[renamed_table] = (a_source, renamed_table)
            source = DuckDBSource(uri=":memory:", mirrors=mirrors)
        else:
            source = next(iter(sources))

        # check whether the SQL query is valid
        expr_slug = output.expr_slug
        try:
            # TODO: if original sql expr matches, don't recreate a new one!
            sql_expr_source = source.create_sql_expr_source({expr_slug: sql_query})
            sql_query = sql_expr_source.tables[expr_slug]
            sql_transforms = [SQLLimit(limit=1_000_000, write=source.dialect, pretty=True, identify=False)]
            transformed_sql_query = sql_query
            for sql_transform in sql_transforms:
                transformed_sql_query = sql_transform.apply(transformed_sql_query)  # not to be used elsewhere; just for transparency
                if transformed_sql_query != sql_query:
                    stream_details(f'```sql\n{transformed_sql_query}\n```', step, title=f"{sql_transform.__class__.__name__} Applied")
            pipeline = await get_pipeline(
                source=sql_expr_source, table=expr_slug, sql_transforms=sql_transforms
            )
        except Exception as e:
            report_error(e, step)
            raise e

        try:
            df = await get_data(pipeline)
        except Exception as e:
            if isinstance(source, DuckDBSource):
                source._connection.execute(f'DROP TABLE IF EXISTS "{expr_slug}"')
            report_error(e, step)
            raise e

        self._memory["data"] = await describe_data(df)
        self._memory["sources"].append(sql_expr_source)
        self._memory["source"] = sql_expr_source
        self._memory["pipeline"] = pipeline
        self._memory["table"] = pipeline.table
        return sql_query

    def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        memory['sql'] = event.new

    @retry_llm_output()
    async def _find_tables(self, messages: list[Message], errors: list | None = None) -> tuple[dict[str, BaseSQLSource], str]:
        sources = {source.name: source for source in self._memory["sources"]}
        vector_metaset = self._memory["sql_metaset"].vector_metaset
        selected_slugs = list(
            #  if no tables/cols are subset
            vector_metaset.selected_columns or
            vector_metaset.vector_metadata_map
        )
        if len(selected_slugs) == 0:
            raise ValueError("No tables found in memory.")

        chain_of_thought = ""
        with self._add_step(title="Determining tables to use", steps_layout=self._steps_layout) as step:
            if len(selected_slugs) > 1:
                system = await self._render_prompt(
                    "find_tables", messages, separator=SOURCE_TABLE_SEPARATOR, errors=errors,
                )
                find_tables_model = self._get_model("find_tables", tables=selected_slugs)
                model_spec = self.prompts["find_tables"].get("llm_spec", self.llm_spec_key)
                response = self.llm.stream(
                    messages,
                    system=system,
                    model_spec=model_spec,
                    response_model=find_tables_model,
                )
                async for output in response:
                    chain_of_thought = output.chain_of_thought or ""
                    if output.potential_join_issues is not None:
                        chain_of_thought += output.potential_join_issues
                    step.stream(chain_of_thought, replace=True)

                if selected_slugs := output.selected_tables:
                    stream_details('\n'.join(slug.split(SOURCE_TABLE_SEPARATOR)[-1] for slug in selected_slugs), step, title="Relevant tables", auto=False)
                    step.success_title = f'Found {len(selected_slugs)} relevant table(s)'

            tables_to_source = {}
            step.stream("Planning to use:")
            for table_slug in selected_slugs:
                a_source_obj, a_table = parse_table_slug(table_slug, sources)
                tables_to_source[a_table] = a_source_obj
                step.stream(f"\n\n{a_table!r}")
        return tables_to_source, chain_of_thought

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        tables_to_source, comments = await self._find_tables(messages)
        dialect = self._memory["source"].dialect
        try:
            sql_query = await self._create_valid_sql(messages, dialect, comments, step_title, tables_to_source)
            pipeline = self._memory['pipeline']
        except RetriesExceededError as e:
            traceback.print_exception(e)
            self._memory["__error__"] = str(e)
            return None
        self._render_lumen(pipeline, spec=sql_query, messages=messages, render_output=render_output, title=step_title)
        return pipeline


class DbtslAgent(LumenBaseAgent, DbtslMixin):
    """
    Responsible for creating and executing queries against a dbt Semantic Layer
    to answer user questions about business metrics.
    """

    conditions = param.List(default=[
        "Always use this when dbtsl_metaset is available",
    ])

    purpose = param.String(default="""
        Responsible for displaying tables to answer user queries about
        business metrics using dbt Semantic Layers. This agent can compile
        and execute metric queries against a dbt Semantic Layer.""")

    prompts = param.Dict(
        default={
            "main": {
                "response_model": DbtslQueryParams,
                "template": PROMPTS_DIR / "DbtslAgent" / "main.jinja2",
            },
            "retry_output": {
                "response_model": RetrySpec,
                "template": PROMPTS_DIR / "LumenBaseAgent" / "retry_output.jinja2"
            },
        }
    )

    provides = param.List(default=["table", "sql", "pipeline", "data", "dbtsl_vector_metaset", "dbtsl_sql_metaset"], readonly=True)

    requires = param.List(default=["source", "dbtsl_metaset"], readonly=True)

    source = param.ClassSelector(class_=BaseSQLSource, doc="""
        The source associated with the dbt Semantic Layer.""")

    _extensions = ('codeeditor', 'tabulator',)

    _output_type = SQLOutput

    def __init__(self, source: Source, **params):
        super().__init__(source=source, **params)

    def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        """
        Update the SQL specification in memory.
        """
        memory["sql"] = event.new

    @retry_llm_output()
    async def _create_valid_query(
        self,
        messages: list[Message],
        title: str | None = None,
        errors: list | None = None
    ):
        """
        Create a valid dbt Semantic Layer query based on user messages.
        """
        system_prompt = await self._render_prompt(
            "main",
            messages,
            errors=errors,
        )

        with self._add_step(title=title or "dbt Semantic Layer query", steps_layout=self._steps_layout) as step:
            model_spec = self.prompts["main"].get("llm_spec", self.llm_spec_key)
            response = self.llm.stream(
                messages,
                system=system_prompt,
                model_spec=model_spec,
                response_model=DbtslQueryParams,
            )

            query_params = None
            try:
                async for output in response:
                    step_message = output.chain_of_thought or ""
                    step.stream(step_message, replace=True)

                query_params = {
                    "metrics": output.metrics,
                    "group_by": output.group_by,
                    "limit": output.limit,
                    "order_by": output.order_by,
                    "where": output.where,
                }
                expr_slug = output.expr_slug

                # Ensure required fields are present
                if "metrics" not in query_params or not query_params["metrics"]:
                    raise ValueError("Query must include at least one metric")

                # Ensure limit exists with a default value if not provided
                if "limit" not in query_params or not query_params["limit"]:
                    query_params["limit"] = 10000

                formatted_params = json.dumps(query_params, indent=2)
                step.stream(f"\n\n`{expr_slug}`\n```json\n{formatted_params}\n```")

                self._memory["dbtsl_query_params"] = query_params
            except asyncio.CancelledError as e:
                step.failed_title = "Cancelled dbt Semantic Layer query generation"
                raise e

        if step.failed_title and step.failed_title.startswith("Cancelled"):
            raise asyncio.CancelledError()
        elif not query_params:
            raise ValueError("No dbt Semantic Layer query was generated.")

        try:
            # Execute the query against the dbt Semantic Layer
            client = self._get_dbtsl_client()
            async with client.session():
                sql_query = await client.compile_sql(
                    metrics=query_params.get('metrics', []),
                    group_by=query_params.get('group_by'),
                    limit=query_params.get('limit'),
                    order_by=query_params.get('order_by'),
                    where=query_params.get('where')
                )

                # Log the compiled SQL for debugging
                step.stream(f"\nCompiled SQL:\n```sql\n{sql_query}\n```", replace=False)

            sql_expr_source = self.source.create_sql_expr_source({expr_slug: sql_query})
            self._memory["sql"] = sql_query

            # Apply transforms
            sql_transforms = [SQLLimit(limit=1_000_000, write=self.source.dialect, pretty=True, identify=False)]
            transformed_sql_query = sql_query
            for sql_transform in sql_transforms:
                transformed_sql_query = sql_transform.apply(transformed_sql_query)
                if transformed_sql_query != sql_query:
                    stream_details(f'```sql\n{transformed_sql_query}\n```', step, title=f"{sql_transform.__class__.__name__} Applied")

            # Create pipeline and get data
            pipeline = await get_pipeline(
                source=sql_expr_source, table=expr_slug, sql_transforms=sql_transforms
            )

            df = await get_data(pipeline)
            sql_metaset = await get_metaset({sql_expr_source.name: sql_expr_source}, [expr_slug])
            vector_metaset = sql_metaset.vector_metaset

            # Update memory
            self._memory["data"] = await describe_data(df)
            self._memory["sources"].append(sql_expr_source)
            self._memory["source"] = sql_expr_source
            self._memory["pipeline"] = pipeline
            self._memory["table"] = pipeline.table
            self._memory["dbtsl_vector_metaset"] = vector_metaset
            self._memory["dbtsl_sql_metaset"] = sql_metaset
            return sql_query, pipeline
        except Exception as e:
            report_error(e, step)
            raise e

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        """
        Responds to user messages by generating and executing a dbt Semantic Layer query.
        """
        try:
            sql_query, pipeline = await self._create_valid_query(messages, step_title)
        except RetriesExceededError as e:
            traceback.print_exception(e)
            self._memory["__error__"] = str(e)
            return None

        self._render_lumen(pipeline, spec=sql_query, messages=messages, render_output=render_output, title=step_title)
        return pipeline



class BaseViewAgent(LumenBaseAgent):

    requires = param.List(default=["pipeline"], readonly=True)

    provides = param.List(default=["view"], readonly=True)

    prompts = param.Dict(
        default={
            "main": {"template": PROMPTS_DIR / "BaseViewAgent" / "main.jinja2"},
        }
    )

    def __init__(self, **params):
        self._last_output = None
        super().__init__(**params)


    @retry_llm_output()
    async def _create_valid_spec(
        self,
        messages: list[Message],
        pipeline: Pipeline,
        schema: dict[str, Any],
        step_title: str | None = None,
        errors: list[str] | None = None,
    ) -> dict[str, Any]:
        errors_context = {}
        if errors:
            errors = ('\n'.join(f"{i+1}. {error}" for i, error in enumerate(errors))).strip()
            try:
                last_output = load_json(self._last_output["json_spec"])
            except Exception:
                last_output = ""

            if "sql_metaset" in self._memory:
                vector_metadata_map = self._memory["sql_metaset"].vector_metaset.vector_metadata_map
            elif "dbtsl_sql_metaset" in self._memory:
                vector_metadata_map = self._memory["dbtsl_sql_metaset"].vector_metaset.vector_metadata_map
            columns_context = ""
            for table_slug, vector_metadata in vector_metadata_map.items():
                table_name = table_slug.split(SOURCE_TABLE_SEPARATOR)[-1]
                if table_name in pipeline.table:
                    columns = [col.name for col in vector_metadata.columns]
                    columns_context += f"\nSQL: {vector_metadata.base_sql}\nColumns: {', '.join(columns)}\n\n"
            errors_context = {
                "errors": errors,
                "last_output": last_output,
                "num_errors": len(errors),
                "columns_context": columns_context,
            }

        doc = self.view_type.__doc__.split("\n\n")[0] if self.view_type.__doc__ else self.view_type.__name__
        system = await self._render_prompt(
            "main",
            messages,
            table=pipeline.table,
            doc=doc,
            **errors_context,
        )

        model_spec = self.prompts["main"].get("llm_spec", self.llm_spec_key)
        response = self.llm.stream(
            messages,
            system=system,
            model_spec=model_spec,
            response_model=self._get_model("main", schema=schema),
        )

        e = None
        error = ""
        with self._add_step(
            title=step_title or "Generating view...",
            steps_layout=self._steps_layout
        ) as step:
            async for output in response:
                chain_of_thought = output.chain_of_thought or ""
                step.stream(chain_of_thought, replace=True)

            self._last_output = dict(output)
            try:
                spec = await self._extract_spec(self._last_output)
            except Exception as e:
                error = str(e)
                traceback.print_exception(e)
                context = f'```\n{yaml.safe_dump(load_json(self._last_output["json_spec"]))}\n```'
                report_error(e, step, language="json", context=context)

        if error:
            raise ValueError(error)

        log_debug(f"{self.name} settled on spec: {spec!r}.")
        return spec

    async def _extract_spec(self, spec: dict[str, Any]):
        return dict(spec)

    async def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        memory['view'] = dict(await self._extract_spec(event.new), type=self.view_type)

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
    ) -> Any:
        """
        Generates a visualization based on user messages and the current data pipeline.
        """
        pipeline = self._memory.get("pipeline")
        if not pipeline:
            raise ValueError("No current pipeline found in memory.")

        schema = await get_schema(pipeline)
        if not schema:
            raise ValueError("Failed to retrieve schema for the current pipeline.")

        spec = await self._create_valid_spec(messages, pipeline, schema, step_title)
        self._memory["view"] = dict(spec, type=self.view_type)
        view = self.view_type(pipeline=pipeline, **spec)
        self._render_lumen(view, messages=messages, render_output=render_output, title=step_title)
        return view


class hvPlotAgent(BaseViewAgent):

    purpose = param.String(default="Generates a plot of the data given a user prompt.")

    prompts = param.Dict(
        default={
            "main": {"template": PROMPTS_DIR / "hvPlotAgent" / "main.jinja2"},
        }
    )

    view_type = hvPlotUIView

    def _get_model(self, prompt_name: str, schema: dict[str, Any]) -> type[BaseModel]:
        # Find parameters
        excluded = self.view_type._internal_params + [
            "controls",
            "type",
            "source",
            "pipeline",
            "transforms",
            "download",
            "field",
            "selection_group",
        ]
        model = param_to_pydantic(self.view_type, base_model=PartialBaseModel, excluded=excluded, schema=schema, extra_fields={
            "chain_of_thought": (str, FieldInfo(description="Your thought process behind the plot.")),
        })
        return model[self.view_type.__name__]

    async def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        spec = yaml.load(event.new, Loader=yaml.SafeLoader)
        memory['view'] = dict(await self._extract_spec(spec), type=self.view_type)

    async def _extract_spec(self, spec: dict[str, Any]):
        pipeline = self._memory["pipeline"]
        spec = {
            key: val for key, val in spec.items()
            if val is not None
        }
        spec["type"] = "hvplot_ui"
        self.view_type.validate(spec)
        spec.pop("type", None)

        # Add defaults
        spec["responsive"] = True
        data = await get_data(pipeline)
        if len(data) > 20000 and spec["kind"] in ("line", "scatter", "points"):
            spec["rasterize"] = True
            spec["cnorm"] = "log"
        return spec


class VegaLiteAgent(BaseViewAgent):

    purpose = param.String(default="Generates a vega-lite specification of the plot the user requested.")

    prompts = param.Dict(
        default={
            "main": {
                "response_model": VegaLiteSpec,
                "template": PROMPTS_DIR / "VegaLiteAgent" / "main.jinja2"
            },
            "retry_output": {
                "response_model": RetrySpec,
                "template": PROMPTS_DIR / "VegaLiteAgent" / "retry_output.jinja2"
            },
        }
    )

    view_type = VegaLiteView

    _extensions = ('vega',)

    _output_type = VegaLiteOutput

    async def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        try:
            spec = await self._extract_spec({"yaml_spec": event.new})
        except Exception as e:
            traceback.print_exception(e)
            return
        memory['view'] = dict(spec, type=self.view_type)

    def _add_geographic_items(self, vega_spec: dict, vega_spec_str: str):
        # standardize the vega spec, by migrating to layer
        if "layer" not in vega_spec:
            vega_spec["layer"] = [{"encoding": vega_spec.pop("encoding", None), "mark": vega_spec.pop("mark", None)}]

        # make it zoomable
        if "params" not in vega_spec:
            vega_spec["params"] = []

        # Get existing param names
        existing_param_names = {
            param.get('name') for param in vega_spec["params"]
            if isinstance(param, dict) and 'name' in param
        }
        for p in VEGA_ZOOMABLE_MAP_ITEMS["params"]:
            if p.get('name') not in existing_param_names:
                vega_spec["params"].append(p)

        if "projection" not in vega_spec:
            vega_spec["projection"] = {"type": "mercator"}
        vega_spec["projection"].update(VEGA_ZOOMABLE_MAP_ITEMS["projection"])

        # Handle map projections and add geographic outlines
        # - albersUsa projection is incompatible with world map data
        # - Each map type needs appropriate boundary outlines
        has_world_map = "world-110m.json" in vega_spec_str
        uses_albers_usa = vega_spec["projection"]["type"] == "albersUsa"

        # If trying to use albersUsa with world map, switch to mercator projection
        if has_world_map and uses_albers_usa:
            vega_spec["projection"] = "mercator"  # cannot use albersUsa with world-110m
        # Add world map outlines if needed
        elif not has_world_map and not uses_albers_usa:
            vega_spec["layer"].append(VEGA_MAP_LAYER["world"])
        return vega_spec

    async def _ensure_columns_exists(self, vega_spec: dict):
        schema = await get_schema(self._memory["pipeline"])

        for layer in vega_spec.get("layer", []):
            encoding = layer.get("encoding", {})
            if not encoding:
                continue

            for enc_def in encoding.values():
                fields_to_check = []

                if isinstance(enc_def, dict) and "field" in enc_def:
                    fields_to_check.append(enc_def["field"])
                elif isinstance(enc_def, list):
                    fields_to_check.extend(
                        item["field"] for item in enc_def
                        if isinstance(item, dict) and "field" in item
                    )

                for field in fields_to_check:
                    if field not in schema and field.lower() not in schema and field.upper() not in schema:
                        raise ValueError(f"Field '{field}' not found in schema.")

    async def _extract_spec(self, spec: dict[str, Any]):
        # .encode().decode('unicode_escape') fixes a JSONDecodeError in Python
        # where it's expecting property names enclosed in double quotes
        # by properly handling the escaped characters in your JSON string
        if yaml_spec:= spec.get('yaml_spec'):
            vega_spec = yaml.load(yaml_spec, Loader=yaml.SafeLoader)
        elif json_spec:= spec.get('json_spec'):
            vega_spec = load_json(json_spec)
        if "$schema" not in vega_spec:
            vega_spec["$schema"] = "https://vega.github.io/schema/vega-lite/v5.json"
        if "width" not in vega_spec:
            vega_spec["width"] = "container"
        if "height" not in vega_spec:
            vega_spec["height"] = "container"
        self._output_type._validate_spec(vega_spec)
        await self._ensure_columns_exists(vega_spec)

        # using string comparison because these keys could be in different nested levels
        vega_spec_str = yaml.dump(vega_spec)
        # Handle different types of interactive controls based on chart type
        if "latitude:" in vega_spec_str or "longitude:" in vega_spec_str:
            vega_spec = self._add_geographic_items(vega_spec, vega_spec_str)
        elif ("point: true" not in vega_spec_str or "params" not in vega_spec) and vega_spec_str.count("encoding:") == 1:
            # add pan/zoom controls to all plots except geographic ones and points overlaid on line plots
            # because those result in an blank plot without error
            vega_spec["params"] = [{"bind": "scales", "name": "grid", "select": "interval"}]
        return {'spec': vega_spec, "sizing_mode": "stretch_both", "min_height": 300, "max_width": 1200}


class AnalysisAgent(LumenBaseAgent):

    analyses = param.List([])

    purpose = param.String(default="Perform custom analyses on the data.")

    prompts = param.Dict(
        default={
            "main": {"template": PROMPTS_DIR / "AnalysisAgent" / "main.jinja2"},
        }
    )

    provides = param.List(default=['view'])

    requires = param.List(default=['pipeline'])

    _output_type = AnalysisOutput

    def _update_spec(self, memory: _Memory, event: param.parameterized.Event):
        pass

    async def respond(
        self,
        messages: list[Message],
        render_output: bool = False,
        step_title: str | None = None,
        agents: list[Agent] | None = None,
    ) -> Any:
        pipeline = self._memory['pipeline']
        analyses = {a.name: a for a in self.analyses if await a.applies(pipeline)}
        if not analyses:
            log_debug("No analyses apply to the current data.")
            return None

        # Short cut analysis selection if there's an exact match
        if len(messages):
            analysis = messages[0].get('content').replace('Apply ', '')
            if analysis in analyses:
                analyses = {analysis: analyses[analysis]}

        if len(analyses) > 1:
            with self._add_step(title="Choosing the most relevant analysis...", steps_layout=self._steps_layout) as step:
                type_ = Literal[tuple(analyses)]
                analysis_model = create_model(
                    "Analysis",
                    correct_name=(type_, FieldInfo(description="The name of the analysis that is most appropriate given the user query."))
                )
                system_prompt = await self._render_prompt(
                    "main",
                    messages,
                    analyses=analyses,
                    data=self._memory.get("data"),
                )
                model_spec = self.prompts["main"].get("llm_spec", self.llm_spec_key)
                analysis_name = (await self.llm.invoke(
                    messages,
                    system=system_prompt,
                    model_spec=model_spec,
                    response_model=analysis_model,
                    allow_partial=False,
                )).correct_name
                step.stream(f"Selected {analysis_name}")
                step.success_title = f"Selected {analysis_name}"
        else:
            analysis_name = next(iter(analyses))

        view = None
        with self.interface.param.update(callback_exception="raise"):
            with self._add_step(title=step_title or "Creating view...", steps_layout=self._steps_layout) as step:
                await asyncio.sleep(0.1)  # necessary to give it time to render before calling sync function...
                analysis_callable = analyses[analysis_name].instance(agents=agents)

                data = await get_data(pipeline)
                for field in analysis_callable._field_params:
                    analysis_callable.param[field].objects = list(data.columns)
                self._memory["analysis"] = analysis_callable

                if analysis_callable.autorun:
                    if asyncio.iscoroutinefunction(analysis_callable.__call__):
                        view = await analysis_callable(pipeline)
                    else:
                        view = await asyncio.to_thread(analysis_callable, pipeline)
                    if isinstance(view, Viewable):
                        view = Panel(object=view, pipeline=self._memory.get('pipeline'))
                    spec = view.to_spec()
                    if isinstance(view, View):
                        view_type = view.view_type
                        self._memory["view"] = dict(spec, type=view_type)
                    elif isinstance(view, Pipeline):
                        self._memory["pipeline"] = view
                    # Ensure data reflects processed pipeline
                    if pipeline is not self._memory['pipeline']:
                        pipeline = self._memory['pipeline']
                        if len(data) > 0:
                            self._memory["data"] = await describe_data(data)
                    yaml_spec = yaml.dump(spec)
                    step.stream(f"Generated view\n```yaml\n{yaml_spec}\n```")
                    step.success_title = "Generated view"
                else:
                    step.success_title = "Configure the analysis"

        analysis = self._memory["analysis"]
        pipeline = self._memory['pipeline']
        if view is None and analysis.autorun:
            self.interface.stream('Failed to find an analysis that applies to this data')
        else:
            self._render_lumen(
                view,
                analysis=analysis,
                pipeline=pipeline,
                render_output=render_output,
                title=step_title
            )
        return view



from __future__ import annotations

import contextlib
import datetime
import decimal
import re

from pathlib import Path
from typing import Any

import pandas as pd
import param
import snowflake.connector

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import (
    Encoding, NoEncryption, PrivateFormat, load_pem_private_key,
)

from ..transforms.sql import SQLFilter
from .base import BaseSQLSource, cached, cached_schema

# PEM certificates have the pattern:
#   -----BEGIN PKEY-----  # renamed private... to pkey to prevent ruff from flagging
#   <- multiple lines of encoded data->
#   -----END PKEY-----
#
# The regex captures the header and footer into groups 1 and 3, the body into group 2
# group 1: "header" captures series of hyphens followed by anything that is
#           not a hyphen followed by another string of hyphens
# group 2: "body" capture everything upto the next hyphen
# group 3: "footer" duplicates group 1
_SIMPLE_PEM_CERTIFICATE_REGEX = "^(-+[^-]+-+)([^-]+)(-+[^-]+-+)"


class SnowflakeSource(BaseSQLSource):
    """
    SnowflakeSource uses the snowflake-python-connector library to load data
    from Snowflake.
    """

    account = param.String(default=None, doc="""
        The account identifier to connect to.""")

    authenticator = param.Selector(default=None, objects=[
        'externalbrowser', 'oauth', 'snowflake', 'username_password_mfa', 'SNOWFLAKE_JWT'], doc="""
        The authentication approach to use.""", allow_None=True)

    conn_kwargs = param.Dict(default={}, doc="""
        Additional connection parameters to pass to the Snowflake connector.""")

    database = param.String(default=None, doc="""
        The database to connect to.""")

    host = param.String(default=None, doc="""
        The host to authenticate with.""")

    token = param.String(default=None, doc="""
        The OAuth token if authenticator is set to "oauth".""")

    user = param.String(default=None, doc="""
        The user to authenticate as.""")

    password = param.String(default=None, doc="""
        The password to authenticate with (if authenticator is set to "snowflake").""")

    private_key = param.ClassSelector(
        default=None, class_=(str, bytes, Path), doc="""
        The path or contents of the private key file.""")

    private_key_password = param.String(default=None, doc="""
        The password to decrypt the private key file.""")

    schema = param.String(default=None, doc="""
        The database schema to load data from.""")

    warehouse = param.String(default=None, doc="""
        The warehouse to connect to.""")

    filter_in_sql = param.Boolean(default=True, doc="""
        Whether to apply filters in SQL or in-memory.""")

    sql_expr = param.String(default='SELECT * FROM {table}', doc="""
        The SQL expression to execute.""")

    tables = param.ClassSelector(class_=(list, dict), doc="""
        List or dictionary of tables.""")

    excluded_tables = param.List(default=[], doc="""
        List of table names that should be excluded from the results.
        The items can be fully qualified (database.schema.table), partially
        qualified (schema.table), simply table names, or wildcards
        (e.g. database.schema.*).""")

    schema_timeout_seconds = param.Integer(default=600, doc="""
        Timeout in seconds for schema retrieval. If None, no timeout is applied.""")

    dialect = 'snowflake'

    def __init__(self, **params):
        conn = params.pop('conn', None)
        super().__init__(**params)
        conn_kwargs = self.conn_kwargs.copy()
        if self.account is not None:
            conn_kwargs['account'] = self.account
        if self.authenticator is not None:
            conn_kwargs['authenticator'] = self.authenticator
        if self.database is not None:
            conn_kwargs['database'] = self.database
        if self.host is not None:
            conn_kwargs['host'] = self.host
        if self.token is not None:
            conn_kwargs['token'] = self.token
        if self.password is not None:
            conn_kwargs['password'] = self.password
        if self.private_key is not None:
            conn_kwargs['private_key'] = self.resolve_private_key()
        if self.schema is not None:
            conn_kwargs['schema'] = self.schema
        if self.user is not None:
            conn_kwargs['user'] = self.user
        if self.warehouse is not None:
            conn_kwargs['warehouse'] = self.warehouse

        self._conn = conn or snowflake.connector.connect(**conn_kwargs)
        self._cursor = self._conn.cursor()

    @staticmethod
    def _decode_secret(secret: str | bytes) -> bytes | None:
        """
        Decode the provided secret into bytes.

        If the secret is not a string or bytes, or it is whitespace, then return None.

        Parameters
        ----------
        secret : str or bytes
            The value to decode.

        Returns
        -------
        bytes or None
            The decoded secret as bytes.
        """
        if not isinstance(secret, (bytes, str)) or len(secret) == 0 or secret.isspace():
            return None

        return secret if isinstance(secret, bytes) else secret.encode()

    @staticmethod
    def _compose_pem(private_key: bytes) -> bytes:
        """
        Validate structure of PEM certificate.

        The original key passed from Prefect is sometimes malformed.
        This function recomposes the key into a valid key that will
        pass the serialization step when resolving the key to a DER.

        Parameters
        ----------
        private_key : bytes
            A valid PEM format byte encoded string.

        Returns
        -------
        bytes
            Byte encoded certificate.

        Raises
        ------
        InvalidPemFormat
            If private key is an invalid format.
        """
        pem_parts = re.match(_SIMPLE_PEM_CERTIFICATE_REGEX, private_key.decode())
        if pem_parts is None:
            raise ValueError("Invalid PEM format")

        body = "\n".join(re.split(r"\s+", pem_parts[2].strip()))
        # reassemble header+body+footer
        return f"{pem_parts[1]}\n{body}\n{pem_parts[3]}".encode()

    def resolve_private_key(self):
        """
        Converts a PEM encoded private key into a DER binary key.

        Parameters
        ----------

        Returns
        -------
        bytes or None
            DER encoded key if private_key has been provided otherwise returns None.

        Raises
        ------
        InvalidPemFormat
            If private key is not in PEM format.
        """
        if Path(self.private_key).exists():
            private_key = Path(self.private_key).read_bytes()
        else:
            private_key = self._decode_secret(self.private_key)

        if self.private_key_password is not None:
            password = self._decode_secret(self.private_key_password)
        else:
            password = None

        composed_private_key = self._compose_pem(private_key)
        return load_pem_private_key(
            data=composed_private_key,
            password=password,
            backend=default_backend(),
        ).private_bytes(
            encoding=Encoding.DER,
            format=PrivateFormat.PKCS8,
            encryption_algorithm=NoEncryption(),
        )

    def create_sql_expr_source(self, tables: dict[str, str], **kwargs):
        """
        Creates a new SQL Source given a set of table names and
        corresponding SQL expressions.
        """
        params = dict(self.param.values(), **kwargs)
        params.pop("name", None)
        params['tables'] = tables
        params['conn'] = self._conn
        return SnowflakeSource(**params)

    def _cast_to_supported_dtypes(self, df: pd.DataFrame, sample: int = 100) -> pd.DataFrame:
        """
        Convert decimal.Decimal to float in a pandas DataFrame, as
        Bokeh ColumnDataSource does not support decimal.Decimal.
        Samples only a subset of the DataFrame to check for decimal.Decimal
        Arguments
        ---------
        df (pd.DataFrame):
        the DataFrame to convert
        sample (int):
        number of rows to sample to check for decimal.Decimal
        """
        df = df.copy()
        for col in df.select_dtypes(include=["object"]).columns:
            df_col_sample = df[col].sample(min(sample, len(df)))
            try:
                if df_col_sample.apply(lambda x: isinstance(x, decimal.Decimal)).any():
                    df[col] = pd.to_numeric(df[col], errors="coerce")
                if df_col_sample.apply(
                    lambda x: isinstance(x, (datetime.datetime, datetime.date))
                ).any():
                    df[col] = pd.to_datetime(df[col], errors="coerce")
            except Exception:
                df[col] = df[col].astype(str)
        return df

    def execute(self, sql_query: str, *args, **kwargs):
        # TODO: remove cast in future, but keep until bokeh has a solution
        df = self._cursor.execute(sql_query, *args, **kwargs).fetch_pandas_all()
        return self._cast_to_supported_dtypes(df)

    def get_tables(self) -> list[str]:
        # limited set of tables was provided
        if isinstance(self.tables, dict | list):
            return [t.upper() for t in list(self.tables) if not self._is_table_excluded(t)]

        tables_df = self.execute(f'SELECT TABLE_NAME, TABLE_SCHEMA FROM {self.database}.INFORMATION_SCHEMA.TABLES;')
        return [
            f'{self.database}.{row.TABLE_SCHEMA}.{row.TABLE_NAME}'.upper()
            for _, row in tables_df.iterrows()
            if not self._is_table_excluded(f'{self.database}.{row.TABLE_SCHEMA}.{row.TABLE_NAME}')
        ]

    @cached
    def get(self, table, **query):
        query.pop('__dask', None)
        sql_expr = self.get_sql_expr(table)
        sql_transforms = query.pop('sql_transforms', [])
        conditions = list(query.items())
        if self.filter_in_sql:
            sql_transforms = [SQLFilter(conditions=conditions)] + sql_transforms
        for st in sql_transforms:
            sql_expr = st.apply(sql_expr)
        return self.execute(sql_expr)

    @contextlib.contextmanager
    def _timeout_context(self, seconds=None):
        """Context manager to apply a timeout for operations using Snowflake's
        ALTER SESSION statement to set a query timeout.

        Parameters
        ----------
        seconds : int or None
            Timeout in seconds. If None, no timeout is applied.
        """
        if seconds is None:
            yield
            return

        try:
            # Set the timeout in seconds
            self._cursor.execute(f"ALTER SESSION SET STATEMENT_TIMEOUT_IN_SECONDS = {seconds}")
            yield
        finally:
            # Restore the original timeout
            self._cursor.execute("ALTER SESSION UNSET STATEMENT_TIMEOUT_IN_SECONDS")

    @cached_schema
    def get_schema(self, table: str | None = None, limit: int | None = None, shuffle: bool = False) -> dict:
        """Returns JSON schema describing the tables returned by the Source.

        Parameters
        ----------
        table : str | None
            The name of the table to return the schema for. If None
            returns schema for all available tables.
        limit : int | None
            Limits the number of rows considered for the schema calculation
        shuffle : bool
            Whether to shuffle the rows when sampling

        Returns
        -------
        dict
            JSON schema(s) for one or all the tables.
        """
        try:
            with self._timeout_context(self.schema_timeout_seconds):
                return super().get_schema(table, limit, shuffle)
        except snowflake.connector.errors.ProgrammingError as e:
            if "exceeded" in str(e) and "timeout" in str(e).lower():
                self.param.warning(f"Schema retrieval timed out: {e}")
            else:
                # Re-raise if it's not a timeout error
                raise
            if table is None:
                return {t: {} for t in self.get_tables()}
            else:
                return {}

    def _get_table_metadata(self, tables: list[str]) -> dict[str, Any]:
        """
        Generate metadata for tables in Snowflake.
        Handles formats: database.schema.table_name, schema.table_name, or table_name.
        Schema can be None to be used as a wildcard.
        """
        def subset_table_slugs(df: pd.DataFrame) -> pd.DataFrame:
            df["TABLE_SLUG"] = df[
                ["TABLE_CATALOG", "TABLE_SCHEMA", "TABLE_NAME"]
            ].agg(lambda x: ".".join(x), axis=1).str.upper()
            df = df[df["TABLE_SLUG"].isin(table_slugs)]
            df = df.drop(
                columns=["TABLE_CATALOG", "TABLE_SCHEMA", "TABLE_NAME"]
            ).set_index("TABLE_SLUG")
            return df

        parsed_tables = []
        for t in tables:
            parts = t.split(".")
            if len(parts) == 3:
                parsed_tables.append(parts)  # database.schema.table_name
            elif len(parts) == 2:
                parsed_tables.append([self.database or "", parts[0], parts[1]])  # schema.table_name
            elif len(parts) == 1:
                parsed_tables.append([self.database or "", self.schema or "", parts[0]])  # table_name
            else:
                raise ValueError(f"Invalid table format: {t}")

        table_slugs = pd.Series([".".join(t) for t in parsed_tables]).str.upper()
        table_metadata = self.execute(
            """
            SELECT
            TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, TABLE_TYPE, COMMENT as TABLE_DESCRIPTION, ROW_COUNT, LAST_ALTERED, CREATED
            FROM INFORMATION_SCHEMA.TABLES
            """
        )
        table_metadata = subset_table_slugs(table_metadata)

        table_cols = self.execute(
            """
            SELECT
            TABLE_CATALOG, TABLE_SCHEMA, TABLE_NAME, COLUMN_NAME, COMMENT AS COLUMN_DESCRIPTION, DATA_TYPE
            FROM INFORMATION_SCHEMA.COLUMNS
            """
        )
        table_cols = subset_table_slugs(table_cols)

        result = {}
        table_metadata_columns = table_metadata.join(table_cols).reset_index()
        for table_slug, group in table_metadata_columns.groupby("TABLE_SLUG"):
            first_row = group.iloc[0]
            description = first_row["TABLE_DESCRIPTION"] or ""
            rows = first_row["ROW_COUNT"]
            rows = None if pd.isna(rows) else int(rows)
            updated_at = first_row["LAST_ALTERED"].isoformat()
            created_at = first_row["CREATED"].isoformat()
            table_type = first_row["TABLE_TYPE"]
            columns = (
                group[["COLUMN_NAME", "COLUMN_DESCRIPTION", "DATA_TYPE"]]
                .rename({"COLUMN_DESCRIPTION": "description", "DATA_TYPE": "data_type"}, axis=1)
                .set_index("COLUMN_NAME")
                .transpose()
                .to_dict()
            )
            result[table_slug] = {
                "description": description,
                "columns": columns,
                "rows": rows,
                "updated_at": updated_at,
                "created_at": created_at,
            }
            if table_type == "VIEW" and "INFORMATION_SCHEMA" not in table_slug:
                try:
                    # DATABASE.SCHEMA.TABLE -> SCHEMA.TABLE
                    table_schema_slug = '.'.join(table_slug.split(".")[1:])
                    result[table_slug]["view_definition"] = self.execute(
                        "SELECT GET_DDL('VIEW', %s)", (table_schema_slug,)
                    ).iloc[0, 0]
                except Exception:
                    result[table_slug]["view_definition"] = "N/A"
        return result



"""Define the package version.

Called __version.py as setuptools_scm will create a _version.py
"""

import os.path

PACKAGE = "lumen"

try:
    # For performance reasons on imports, avoid importing setuptools_scm
    # if not in a .git folder
    if os.path.exists(os.path.join(os.path.dirname(__file__), "..", ".git")):
        # If setuptools_scm is installed (e.g. in a development environment with
        # an editable install), then use it to determine the version dynamically.
        from setuptools_scm import get_version

        # This will fail with LookupError if the package is not installed in
        # editable mode or if Git is not installed.
        __version__ = get_version(root="..", relative_to=__file__)
    else:
        raise FileNotFoundError
except (ImportError, LookupError, FileNotFoundError):
    # As a fallback, use the version that is hard-coded in the file.
    try:
        # __version__ was added in _version in setuptools-scm 7.0.0, we rely on
        # the hopefully stable version variable.
        from ._version import version as __version__
    except (ModuleNotFoundError, ImportError):
        # Either _version doesn't exist (ModuleNotFoundError) or version isn't
        # in _version (ImportError). ModuleNotFoundError is a subclass of
        # ImportError, let's be explicit anyway.

        # Try something else:
        from importlib.metadata import PackageNotFoundError, version

        try:
            __version__ = version(PACKAGE)
        except PackageNotFoundError:
            # The user is probably trying to run this without having installed
            # the package.
            __version__ = "0.0.0+unknown"

__all__ = ("__version__",)



from __future__ import annotations

import datetime as dt
import pathlib
import re

from copy import deepcopy
from typing import ClassVar

import param  # type: ignore
import sqlglot

from sqlglot import parse
from sqlglot.expressions import (
    LT, Column, Expression, Identifier, Literal as SQLLiteral, Max, Min, Null,
    ReadCSV, Select, Star, Table, TableSample, and_, func, or_,
    replace_placeholders, replace_tables, select,
)
from sqlglot.optimizer import optimize

from ..config import SOURCE_TABLE_SEPARATOR
from ..util import detect_file_encoding
from .base import Transform


class SQLTransform(Transform):
    """
    Base class for SQL transforms using sqlglot.
    """

    comments = param.Boolean(
        default=False, doc="Whether to include comments in the output SQL")

    error_level = param.ClassSelector(
        class_=sqlglot.ErrorLevel,
        default=sqlglot.ErrorLevel.RAISE,
        doc="Error level for parsing",
    )

    identify = param.Boolean(default=False, doc="""
        Delimit all identifiers, e.g. turn `FROM database.table` into `FROM "database"."table"`.
        This is useful for dialects that don't support unquoted identifiers.""")

    optimize = param.Boolean(
        default=False, doc="""
        Whether to optimize the generated SQL query; may produce invalid results, especially with
        duckdb's read_* functions.""")

    pretty = param.Boolean(default=False, doc="Prettify output SQL, i.e. add newlines and indentation")

    read = param.String(default=None, doc="Source dialect for parsing; if None, automatically detects")

    unsupported_level = param.ClassSelector(
        class_=sqlglot.ErrorLevel,
        default=sqlglot.ErrorLevel.WARN,
        doc="When using `to_sql`, how to handle unsupported dialect features.",
    )

    write = param.String(
        default=None, doc="Target dialect for output; if None, defaults to read dialect"
    )

    __abstract = True

    def __init__(self, **params):
        for key in ["read", "write"]:
            if params.get(key) == "any":
                params[key] = None

        if params.get("read") and not params.get("write"):
            params["write"] = params["read"]
        elif not params.get("read") and params.get("write"):
            params["read"] = params["write"]
        super().__init__(**params)

    @classmethod
    def apply_to(cls, sql_in: str, **kwargs) -> str:
        """
        Calls the apply method based on keyword arguments passed to define transform.

        Parameters
        ----------
        sql_in: SQL Select statement to input to transformation.

        Returns
        -------
        SQL statement after application of transformation.
        """
        return cls(**kwargs).apply(sql_in)

    def apply(self, sql_in: str) -> str:
        """
        Given an SQL statement, manipulate it, and return a new SQL statement.

        Parameters
        ----------
        sql_in: string
            The initial SQL query to be manipulated.

        Returns
        -------
        string
            New SQL query derived from the above query.
        """
        return self.to_sql(self.parse_sql(sql_in))

    def parse_sql(self, sql_in: str) -> Expression:
        """
        Parse SQL string into sqlglot AST.

        Parameters
        ----------
        sql_in: string
            SQL string to parse

        Returns
        -------
        sqlglot.Expression
            Parsed SQL expression
        """
        try:
            expressions = parse(
                sql_in,
                read=self.read,
                error_level=self.error_level,
            )
        except sqlglot.ParseError:
            # Find table name after FROM and quote it if it contains special characters
            pattern = r'FROM\s+([a-zA-Z0-9_\-\.]+)'
            quoted_sql_in = re.sub(
                pattern, lambda m: f'FROM "{m.group(1)}"', sql_in, flags=re.IGNORECASE
            )
            expressions = parse(
                quoted_sql_in,
                read=self.read,
                error_level=self.error_level,
            )

        if len(expressions) > 1:
            raise ValueError(
                "Multiple SQL statements found. Please provide only a single SQL statement."
            )
        expression = expressions[0]
        return expression

    def _add_encoding_to_read_csv(self, expression: Expression) -> Expression:
        """
        Add file encoding when reading CSV files using DuckDB.

        Parameters
        ----------
        expression : Expression
            An sqlglot expression object.

        Returns
        -------
        Expression
            A modified expression that includes the file encoding.
        """
        expr = deepcopy(expression)
        if isinstance(expr, ReadCSV):
            read_csv = expr.find(ReadCSV) or ReadCSV()
            literal = read_csv.find(SQLLiteral) or SQLLiteral()
            if pathlib.Path(literal.this).suffix.lower() == ".csv" and "encoding" not in literal.this:
                encoding = detect_file_encoding(file_obj=literal.this)
                expr.find(ReadCSV).find(SQLLiteral).replace(Identifier(this=f"'{literal.this}', encoding='{encoding}'", is_string=literal.is_string))

        return expr

    def to_sql(self, expression: Expression) -> str:
        """
        Convert sqlglot expression back to SQL string.

        Parameters
        ----------
        expression: sqlglot.Expression
            Expression to convert to SQL

        Returns
        -------
        string
            SQL string representation
        """
        if self.optimize:
            expression = optimize(expression, dialect=self.read)

        expression = self._add_encoding_to_read_csv(expression=expression)

        return expression.sql(
            comments=self.comments,
            dialect=self.write,
            identify=self.identify,
            pretty=self.pretty,
            unsupported_level=self.unsupported_level,
        )


class SQLFormat(SQLTransform):
    """
    Format SQL expressions with parameterized replacements.

    This transform allows for replacing placeholders in SQL queries using
    either Python string format-style placeholders {name} or sqlglot-style
    placeholders :name.

    Parameters
    ----------
    sql_expr : str
        The SQL expression template with placeholders to be formatted.
    parameters : dict
        Dictionary of parameter names and values to replace in the SQL template.
    py_format : bool
        Whether to convert Python format-style placeholders {name} to sqlglot
        placeholders :name before parsing. Default is True.
    """

    parameters = param.Dict(default={}, doc="""
        Dictionary of parameter names and values to replace in the SQL template.""")

    transform_type: ClassVar[str] = 'sql_format'

    def apply(self, sql_in: str) -> str:
        """
        Apply the formatting to the input SQL, replacing placeholders with values.

        Parameters
        ----------
        sql_in : str
            The input SQL query to format. This is used as a base query that will
            have the formatted sql_expr applied to it, typically as a subquery.

        Returns
        -------
        str
            The formatted SQL query with all placeholders replaced.
        """
        sql_template = re.sub(r'\{(\w+)\}', r':\1', sql_in)
        expression = self.parse_sql(sql_template)
        if self.parameters:
            parameters = {}
            for k, v in self.parameters.items():
                if isinstance(v, str):
                    parameters[k] = Identifier(this=v, quoted=self.identify)
                else:
                    parameters[k] = v
            replaced_expression = replace_placeholders(expression, **parameters)
        return self.to_sql(replaced_expression,)


class SQLSelectFrom(SQLFormat):

    sql_expr = param.String(
        default="SELECT * FROM {table}",
        doc="""
        The SQL expression to useif the sql_in does NOT
        already contain a SELECT statement.""",
    )

    tables = param.ClassSelector(
        default=None,
        class_=(list, dict),
        doc="""
        Dictionary of tables to replace or use in the SQL expression.
        If None, the original table will be used.""",
    )

    transform_type: ClassVar[str] = "sql_select_from"

    def apply(self, sql_in: str) -> str:
        try:
            expression = self.parse_sql(sql_in)
        except sqlglot.ParseError:
            # The expression has no SELECT statement, and invalid characters
            # e.g. read_parquet("/path/to/file.parquet"); so we need to quote
            expression = Table(this=Identifier(this=sql_in, quoted=True))

        # if 'data/life-expectancy.csv' becomes 'data / life-expectancy.csv'
        if not list(expression.find_all(Select)) and " / " in expression.sql():
            expression = Table(this=Identifier(this=sql_in, quoted=True))

        tables = {}

        # one word table names may be classified as a Column
        existing_tables = list(expression.find_all(Table)) or expression.find_all(
            Column
        )
        if isinstance(self.tables, list):
            for existing_table, renamed_table in zip(existing_tables, self.tables):
                if isinstance(renamed_table, Table):
                    tables[existing_table] = renamed_table
                else:
                    tables[existing_table] = Table(
                        this=Identifier(this=renamed_table),
                        alias=Identifier(this=existing_table.alias),
                    )

        else:
            for renamed_key, renamed_value in (self.tables or {}).items():
                if isinstance(renamed_key, Table) and isinstance(renamed_value, Table):
                    tables[renamed_key] = renamed_value
                elif isinstance(renamed_key, Table):
                    tables[renamed_key] = Table(
                        Identifier(this=renamed_value),
                        alias=Identifier(this=renamed_key.alias),
                    )
                elif isinstance(renamed_value, Table):
                    tables[
                        Table(
                            this=renamed_key, alias=Identifier(this=renamed_value.alias)
                        )
                    ] = renamed_value
                else:
                    alias = next(
                        (
                            existing_table.alias
                            for existing_table in existing_tables
                            if renamed_key == str(existing_table.this)
                        ),
                        "",
                    )
                    tables[
                        Table(
                            this=Identifier(this=renamed_key),
                            alias=Identifier(this=alias),
                        )
                    ] = Table(
                        this=Identifier(this=renamed_value),
                        alias=Identifier(this=alias),
                    )

        if expression.find(Select) is not None:
            # if no tables to replace, just return the SQL
            if not tables:
                return sql_in
            # if Select is found, replace tables
            replaced_expression = replace_tables(expression, tables, dialect=self.read)
            return self.to_sql(replaced_expression)

        # if Select is NOT found, use the default sql_expr
        # and if tables are provided, use that table
        # else use the sql_in turned into expression (which should be a table)
        table = next(iter(tables.values()), expression)
        self.parameters = {"table": table}
        return super().apply(self.sql_expr)


class SQLGroupBy(SQLTransform):
    """
    Performs a Group-By and aggregation
    """

    by = param.List(doc="Columns to group by.")

    aggregates = param.Dict(doc="""
        Mapping of aggregate functions to use to which column(s) to use them on,
        e.g. {"AVG": "col1", "SUM": ["col1", "col2"]}.""")

    transform_type: ClassVar[str] = 'sql_group_by'

    def apply(self, sql_in: str) -> str:
        if not self.aggregates:
            raise ValueError(
                'SQLGroupBy transform requires at least one aggregate function to be specified.'
            )
        elif not self.by:
            return sql_in

        subquery = self.parse_sql(sql_in).subquery()
        aggs = []
        for agg, cols in self.aggregates.items():
            for col in [cols] if isinstance(cols, str) else cols:
                aggs.append(f'{agg}({col}) AS {col}')
        expression = select(*self.by, *aggs).from_(subquery).group_by(*self.by)
        return self.to_sql(expression)



class SQLLimit(SQLTransform):
    """
    Performs a LIMIT SQL operation on the query.
    If the query already has a LIMIT clause, it will only be applied if the
    existing limit is less than the new limit.
    """

    limit = param.Integer(default=1000, allow_None=True, doc="Limit on the number of rows to return")

    transform_type: ClassVar[str] = 'sql_limit'

    def apply(self, sql_in: str) -> str:
        if self.limit is None:
            return sql_in

        parsed_expression = self.parse_sql(sql_in)
        if existing_limit := parsed_expression.args.get("limit"):
            if int(existing_limit.expression.this) < self.limit:
                # if existing limit is less than the new limit
                # do not modify the original query
                return sql_in

        subquery = parsed_expression.subquery()
        expression = select("*").from_(subquery).limit(self.limit)
        return self.to_sql(expression)


class SQLDistinct(SQLTransform):

    columns = param.List(default=[], doc="Columns to return distinct values for.")

    transform_type: ClassVar[str] = 'sql_distinct'

    def apply(self, sql_in: str) -> str:
        if not self.columns:
            return sql_in

        subquery = self.parse_sql(sql_in).subquery()
        expression = select(*[Identifier(this=col, quoted=True) for col in self.columns]).from_(subquery).distinct()
        return self.to_sql(expression)


class SQLCount(SQLTransform):

    transform_type: ClassVar[str] = 'sql_count'

    def apply(self, sql_in: str) -> str:
        subquery = self.parse_sql(sql_in).subquery()
        expression = select("COUNT(*) as count").from_(subquery)
        return self.to_sql(expression)


class SQLMinMax(SQLTransform):

    columns = param.List(default=[], doc="Columns to return min/max values for.")

    transform_type: ClassVar[str] = 'sql_minmax'

    def apply(self, sql_in: str) -> str:
        if not self.columns:
            return sql_in

        subquery = self.parse_sql(sql_in).subquery()
        minmax = []
        for col in self.columns:
            quoted = self.identify or bool(re.search(r'\W', col))
            for agg_func in (Min, Max):
                alias = Identifier(this=(col + f"_{agg_func.__name__.lower()}"), quoted=quoted)
                minmax.append(
                    agg_func(this=Column(this=Identifier(this=col, quoted=quoted))).as_(alias)
                )
        expression = select(*minmax).from_(subquery)
        return self.to_sql(expression)


class SQLColumns(SQLTransform):

    columns = param.List(default=[], doc="Columns to return.")

    transform_type: ClassVar[str] = 'sql_columns'

    def apply(self, sql_in: str) -> str:
        if not self.columns:
            return sql_in

        subquery = self.parse_sql(sql_in).subquery()
        expression = select(*self.columns).from_(subquery)
        return self.to_sql(expression)


class SQLFilter(SQLTransform):

    conditions = param.List(doc="""
      List of filter conditions expressed as tuples of the column
      name and the filter value.""")

    transform_type: ClassVar[str] = 'sql_filter'

    def apply(self, sql_in: str) -> str:
        if not self.conditions:
            return sql_in

        subquery = self.parse_sql(sql_in).subquery()
        filters = []

        for col, val in self.conditions:
            column_expr = Column(this=col)

            if val is None:
                filters.append(column_expr.is_(Null()))
            elif isinstance(val, (int, float)):
                filters.append(column_expr.eq(SQLLiteral.number(val)))
            elif isinstance(val, str):
                filters.append(column_expr.eq(SQLLiteral.string(val)))
            elif isinstance(val, dt.datetime):
                filters.append(column_expr.eq(SQLLiteral.string(str(val))))
            elif isinstance(val, dt.date):
                start = SQLLiteral.string(f"{val} 00:00:00")
                end = SQLLiteral.string(f"{val} 23:59:59")
                filters.append(column_expr.between(start, end))
            elif isinstance(val, tuple) and len(val) == 2:
                start, end = val
                if isinstance(start, dt.date) and not isinstance(start, dt.datetime):
                    start_str = f"{start} 00:00:00"
                else:
                    start_str = str(start)

                if isinstance(end, dt.date) and not isinstance(end, dt.datetime):
                    end_str = f"{end} 23:59:59"
                else:
                    end_str = str(end)

                filters.append(column_expr.between(SQLLiteral.string(start_str), SQLLiteral.string(end_str)))
            elif isinstance(val, list):
                if all(isinstance(v, tuple) and len(v) == 2 for v in val):
                    range_filters = []
                    for v1, v2 in val:
                        if isinstance(v1, dt.date) and not isinstance(v1, dt.datetime):
                            v1_str = f"{v1} 00:00:00"
                        else:
                            v1_str = str(v1)

                        if isinstance(v2, dt.date) and not isinstance(v2, dt.datetime):
                            v2_str = f"{v2} 23:59:59"
                        else:
                            v2_str = str(v2)

                        range_filters.append(column_expr.between(SQLLiteral.string(v1_str), SQLLiteral.string(v2_str)))
                    filters.append(or_(*range_filters))
                else:
                    # Handle None values separately in lists
                    non_none_values = [v for v in val if v is not None]
                    has_none = any(v is None for v in val)

                    if has_none and not non_none_values:
                        # All values are None
                        filters.append(column_expr.is_(Null()))
                    elif has_none:
                        # Mix of None and non-None values
                        none_filter = column_expr.is_(Null())
                        in_filter = column_expr.isin(*[SQLLiteral.string(str(v)) for v in non_none_values])
                        filters.append(or_(none_filter, in_filter))
                    else:
                        # No None values
                        filters.append(column_expr.isin(*[SQLLiteral.string(str(v)) for v in val]))
            else:
                self.param.warning(f"Condition {val!r} on {col!r} column not understood. Filter query will not be applied.")
                continue

        expression = select("*").from_(subquery).where(and_(*filters))
        return self.to_sql(expression)


class SQLOverride(SQLTransform):

    override = param.String()

    def apply(self, sql_in: str) -> str:
        return self.override


class SQLSample(SQLTransform):
    """
    Samples rows from a SQL query using TABLESAMPLE or similar functionality,
    depending on the dialect's support.
    """

    percent = param.Number(default=10.0, bounds=(0.0, 100.0), doc="""
        percent of rows to sample. Must be between 0 and 100.""")

    seed = param.Integer(default=None, allow_None=True, doc="""
        Random seed for reproducible sampling.""")

    size = param.Integer(default=None, allow_None=True, doc="""
        Absolute number of rows to sample. If specified, takes precedence over percent.""")

    sample_kwargs = param.Dict(default={}, doc="""
        Other keyword arguments, like method, bucket_numerator, bucket_denominator, bucket_field.""")

    transform_type: ClassVar[str] = 'sql_sample'

    def apply(self, sql_in: str) -> str:
        if (self.percent <= 0 or self.percent >= 100) and self.size is None:
            return sql_in

        expression = self.parse_sql(sql_in)
        dialect = self.write or self.read

        if dialect in ('sqlserver', 'duckdb') or (dialect in ('postgres', 'redshift', 'snowflake', 'bigquery')):
            return self._apply_tablesample_dialect(expression)
        elif dialect in ('mysql', 'mariadb'):
            return self._apply_mysql_dialect(expression)
        elif dialect == 'sqlite':
            return self._apply_sqlite_dialect(expression)
        else:
            return self._apply_generic_dialect(expression)

    def _apply_tablesample_dialect(self, expression: Expression) -> str:
        subquery = expression.subquery("subquery")

        if self.size is not None:
            sample_clause = TableSample(size=SQLLiteral(this=str(self.size), is_string=False))
        else:
            sample_clause = TableSample(percent=SQLLiteral(this=str(self.percent), is_string=False))

        if self.seed is not None:
            sample_clause.args["seed"] = SQLLiteral(this=str(self.seed), is_string=False)

        if "method" in self.sample_kwargs:
            sample_clause.args["method"] = self.sample_kwargs["method"]

        if "bucket_numerator" in self.sample_kwargs:
            sample_clause.args["bucket_numerator"] = SQLLiteral(this=str(self.sample_kwargs["bucket_numerator"]), is_string=False)

        if "bucket_denominator" in self.sample_kwargs:
            sample_clause.args["bucket_denominator"] = SQLLiteral(this=str(self.sample_kwargs["bucket_denominator"]), is_string=False)

        if "bucket_field" in self.sample_kwargs:
            sample_clause.args["bucket_field"] = Column(this=self.sample_kwargs["bucket_field"], is_string=True)

        sample_expression = Select(
            sample=sample_clause,
        ).select("*").from_(subquery)
        return self.to_sql(sample_expression)

    def _apply_mysql_dialect(self, expression: Expression) -> str:
        """Handle MySQL sampling using ORDER BY RAND() with an indexed random column."""
        subquery = expression.subquery("subquery")

        if self.size is not None:
            rand_low = func("RAND")
            rand_high = rand_low + SQLLiteral.number(3 * self.size / 1000000)

            sample_expression = (
                select("*")
                .from_(subquery)
                .where(func("frozen_rand").between(rand_low, rand_high))
                .order_by(func("RAND"))
                .limit(self.size)
            )
        else:
            sample_expression = (
                select("*")
                .from_(subquery)
                .order_by(func("RAND"))
                .limit(
                    func("FLOOR",
                        func("COUNT", Star()) * (self.percent / 100.0)
                    )
                )
            )

        return self.to_sql(sample_expression)

    def _apply_sqlite_dialect(self, expression: Expression) -> str:
        """Handle SQLite sampling which doesn't have native TABLESAMPLE support.

        Uses a more efficient approach with a subquery for sample size-based sampling.
        For percent-based sampling, uses RANDOM() < X filter.
        """
        main_subquery = expression.subquery("main_query")
        sampling_subquery = expression.subquery("sampling_query")
        if self.seed is not None:
            random_func = func("RANDOM", SQLLiteral.number(self.seed))
        else:
            random_func = func("RANDOM")
        if self.size is not None:
            inner_query = (
                select(Column(this="rowid"))
                .from_(sampling_subquery)
                .order_by(random_func)
                .limit(self.size)
            )
            sample_expr = (
                select("*")
                .from_(main_subquery)
                .where(Column(this="rowid").isin(inner_query))
            )
            return self.to_sql(sample_expr)
        else:
            filter_expr = LT(
                this=random_func,
                expression=SQLLiteral.number(self.percent/100.0)
            )
            sample_expr = (
                select("*")
                .from_(main_subquery)
                .where(filter_expr)
            )
            return self.to_sql(sample_expr)

    def _apply_generic_dialect(self, expression: Expression) -> str:
        """Handle sampling for generic SQL dialects."""
        subquery = expression.subquery("subquery")

        if self.seed is not None:
            rand_func = func("RAND", SQLLiteral.number(self.seed))
        else:
            rand_func = func("RAND")

        if self.size is not None:
            sample_expression = (
                select("*")
                .from_(subquery)
                .order_by(rand_func)
                .limit(self.size)
            )
        else:
            filter_expr = LT(
                this=rand_func,
                expression=SQLLiteral.number(self.percent / 100.0)
            )
            sample_expression = (
                select("*")
                .from_(subquery)
                .where(filter_expr)
            )

        return self.to_sql(sample_expression)


class SQLRemoveSourceSeparator(SQLTransform):
    """
    Class to exclude the source and separator.
    """

    separator = param.String(default=SOURCE_TABLE_SEPARATOR, doc="""
        Separator used to split the source and table name in the SQL query.""")

    _parsable_separator = "___AT___"

    def apply(self, sql_in: str) -> str:
        """
        Exclude the source and separator from the SQL query.

        Parameters
        ----------
        sql_in: string
            The initial SQL query to be manipulated.

        Returns
        -------
        string
            New SQL query derived from the above query.
        """
        if self.separator not in sql_in:
            return sql_in
        sql_in = sql_in.replace(SOURCE_TABLE_SEPARATOR, self._parsable_separator)
        return self.to_sql(self.parse_sql(sql_in).transform(self._remove_source_separator))

    def _remove_source_separator(self, node):
        if isinstance(node, Table):
            node_str = node.this
            while hasattr(node_str, "this"):
                node_str = node_str.this
            after = node_str.split(self._parsable_separator)[-1]
            filename = node.this.sql().replace(node_str, after)
            return filename
        elif isinstance(node, sqlglot.exp.Literal) and isinstance(node.this, str):
            filename = node.this.split(self._parsable_separator)[-1]
            return filename
        else:
            return node


__all__ = [name for name, obj in locals().items() if isinstance(obj, type) and issubclass(obj, SQLTransform)]



from __future__ import annotations

import asyncio
import html
import inspect
import json
import math
import re
import time
import traceback

from functools import wraps
from pathlib import Path
from shutil import get_terminal_size
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Literal
from urllib.parse import parse_qs

import pandas as pd

from jinja2 import (
    ChoiceLoader, DictLoader, Environment, FileSystemLoader, StrictUndefined,
)
from markupsafe import escape

from lumen.pipeline import Pipeline

from ..transforms import SQLRemoveSourceSeparator
from ..util import log
from .components import Details
from .config import (
    PROMPTS_DIR, SOURCE_TABLE_SEPARATOR, UNRECOVERABLE_ERRORS,
    RetriesExceededError,
)

if TYPE_CHECKING:
    from panel.chat.step import ChatStep

    from lumen.sources.base import Source


def format_float(num):
    if pd.isna(num) or math.isinf(num):
        return num
    # if is integer, round to 0 decimals
    if num == int(num):
        return f"{int(num)}"
    elif 0.01 <= abs(num) < 100:
        return f"{num:.1f}"  # Regular floating-point notation with one decimal
    else:
        return f"{num:.1e}"  # Exponential notation with one decimal


def fuse_messages(messages: list[dict], max_user_messages: int = 2) -> list[dict]:
    """
    Fuses the chat history into a single system message, followed by the last user message.
    This function is reusable across different components that need to combine
    multiple messages into a simplified format for LLM processing.

    Parameters
    ----------
    messages : list[dict]
        List of message dictionaries with 'role' and 'content' keys
    max_user_messages : int
        Maximum number of user messages to include in the history

    Returns
    -------
    list[dict]
        Processed messages with the chat history as a system message
    """
    user_indices = [i for i, msg in enumerate(messages) if msg['role'] == 'user']
    if user_indices:
        first_user_index = user_indices[0] if len(user_indices) <= max_user_messages else user_indices[-max_user_messages]
        last_user_index = user_indices[-1]
        last_user_message = messages[last_user_index]
    else:
        first_user_index = 0
        last_user_index = -1
    history_messages = messages[first_user_index:last_user_index]
    if not history_messages:
        return [last_user_message] if user_indices else []
    formatted_history = "\n\n".join(
        f"{msg['role'].capitalize()}: {msg['content']}" for msg in history_messages if msg['content'].strip()
    )
    system_prompt = {
        "role": "system",
        "content": f"<Chat History>\n{formatted_history}\n<\\Chat History>"
    }
    return [system_prompt] if last_user_index == -1 else [system_prompt, last_user_message]


def render_template(template_path: Path | str, overrides: dict | None = None, relative_to: Path = PROMPTS_DIR, **context):
    if isinstance(template_path, str):
        template_path = Path(template_path)

    template_name = None

    # Create a list of search paths, starting with the default
    search_paths = [relative_to]
    # If template_path is absolute, add its parent directory to search paths
    if template_path.is_absolute():
        # Add the parent directory of the template to the search paths
        search_paths.append(template_path.parent)
        template_name = template_path.name
    else:
        try:
            # Try to make it relative to the default template directory
            template_name = template_path.relative_to(relative_to).as_posix()
        except ValueError:
            # If that fails, treat it as a plain file path
            # Add parent dir to search paths and use filename as template_name
            if template_path.exists():
                search_paths.append(template_path.parent)
                template_name = template_path.name
            else:
                # Final fallback - assume it's a relative path but not to PROMPTS_DIR
                template_name = str(template_path)

    # Create a loader that can search in multiple directories
    fs_loader = FileSystemLoader(search_paths)

    if overrides:
        # Dynamically create block definitions based on dictionary keys with proper escaping
        block_definitions = "\n".join(
            f"{{% block {escape(key)} %}}{escape(value)}{{% endblock %}}"
            for key, value in overrides.items()
        )
        # Create the dynamic template content by extending the base template and adding blocks
        dynamic_template_content = dedent(
            f"""
            {{% extends "{template_name}" %}}
            {block_definitions}
            """
        )
        dynamic_loader = DictLoader({"dynamic_template": dynamic_template_content})
        choice_loader = ChoiceLoader([dynamic_loader, fs_loader])
        env = Environment(loader=choice_loader, undefined=StrictUndefined)
        template_name = "dynamic_template"
    else:
        env = Environment(loader=fs_loader, undefined=StrictUndefined)

    template = env.get_template(template_name)
    return template.render(**context)


def warn_on_unused_variables(string, kwargs, prompt_label):
    used_keys = set()

    for key in kwargs:
        pattern = r'\b' + re.escape(key) + r'\b'
        if re.search(pattern, string):
            used_keys.add(key)

    unused_keys = set(kwargs.keys()) - used_keys
    if unused_keys:
        # TODO: reword this concisely... what do you call those variables for formatting?
        log.warning(
            f"The prompt template, {prompt_label}, is missing keys, "
            f"which could mean the LLM is lacking the context provided "
            f"from these variables: {unused_keys}. If this is unintended, "
            f"please create a template that contains those keys."
        )


def retry_llm_output(retries=3, sleep=1):
    """
    Retry a function that returns a response from the LLM API.
    If the function raises an exception, retry it up to `retries` times.
    If `sleep` is provided, wait that many seconds between retries.
    If an error occurs, pass the error message into kwargs for the function
    to manually handle by including it.
    """

    def decorator(func):
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args, **kwargs):
                errors = []
                for i in range(retries):
                    if i > 0:
                        log_debug(f"Retrying LLM function {func.__name__} ({i} / {retries})...")
                    if errors:
                        kwargs["errors"] = errors
                    try:
                        output = await func(*args, **kwargs)
                        if not output:
                            raise Exception("No valid output from LLM.")
                        return output
                    except Exception as e:
                        if isinstance(e, UNRECOVERABLE_ERRORS):
                            log_debug(f"LLM encountered unrecoverable error: {e}")
                            raise e
                        elif i == retries - 1:
                            raise RetriesExceededError("Maximum number of retries exceeded.") from e
                        errors.append(str(e))
                        traceback.print_exc()
                        if sleep:
                            await asyncio.sleep(sleep)

            return async_wrapper
        else:

            @wraps(func)
            def sync_wrapper(*args, **kwargs):
                errors = []
                for i in range(retries):
                    if i > 0:
                        log_debug(f"Retrying LLM function {func.__name__} ({i} / {retries})...")
                    if errors:
                        kwargs["errors"] = errors
                    try:
                        output = func(*args, **kwargs)
                        if not output:
                            raise Exception("No valid output from LLM.")
                        return output
                    except Exception as e:
                        if isinstance(e, UNRECOVERABLE_ERRORS):
                            log_debug(f"LLM encountered unrecoverable error: {e}")
                            raise e
                        elif i == retries - 1:
                            raise RetriesExceededError("Maximum number of retries exceeded.") from e
                        errors.append(str(e))
                        traceback.print_exc()
                        if sleep:
                            time.sleep(sleep)

            return sync_wrapper

    return decorator


def process_enums(spec, num_cols, limit=None, include_enum=True, reduce_enums=True, include_empty_fields=False):
    """
    Process enum values in a schema field specification.

    Parameters
    ----------
    spec : dict
        The specification dictionary for this field
    num_cols : int
        The total number of columns in the schema (used for enum reduction calculation)
    limit : int, optional
        The limit used in schema generation
    include_enum : bool, default True
        Whether to include enum values
    reduce_enums : bool, default True
        Whether to reduce the number of enum values for readability
    include_empty_fields : bool, default False
        Whether to include fields that are empty

    Returns
    -------
    tuple[dict, bool]
        The updated spec dictionary and a boolean indicating if this field is empty
    """
    is_empty = False

    if "enum" not in spec:
        return spec, is_empty

    # scale the number of enums based on the number of columns
    if reduce_enums:
        max_enums = max(2, min(10, int(10 * math.exp(-0.1 * max(0, num_cols - 10)))))
    else:
        max_enums = 6
    truncate_limit = min(limit or 5, max_enums)

    if not include_enum or len(spec["enum"]) == 0:
        spec.pop("enum")
    elif len(spec["enum"]) > max_enums:
        # if there are only 10 enums, fine; let's keep them all
        # else, that assumes this column is like glasbey 100+ categories vs category10
        # so truncate to 5 and add "..."
        spec["enum"] = spec["enum"][:truncate_limit] + ["..."]
    elif limit and len(spec["enum"]) == 1 and spec["enum"][0] is None:
        spec["enum"] = [f"(unknown; truncated to {limit} rows)"]
    elif not include_empty_fields and len(spec["enum"]) == 1 and (
        spec["enum"][0] is None or (isinstance(spec["enum"][0], str) and spec["enum"][0].strip() == "")
    ):
        is_empty = True

    # truncate each enum to 100 characters if they exist
    if "enum" in spec:
        spec["enum"] = [
            enum if enum is None or not isinstance(enum, str) or len(enum) < 100 else f"{enum[:100]} ..."
            for enum in spec["enum"]
        ]

    return spec, is_empty


async def get_schema(
    source: Source | Pipeline,
    table: str | None = None,
    include_type: bool = True,
    include_min_max: bool = True,
    include_enum: bool = True,
    include_count: bool = False,
    include_empty_fields: bool = False,
    include_nans: bool = False,
    reduce_enums: bool = True,
    shuffle: bool = True,
    **get_kwargs
):
    if isinstance(source, Pipeline):
        schema = await asyncio.to_thread(source.get_schema)
    else:
        if "limit" not in get_kwargs:
            get_kwargs["limit"] = 100
        schema = await asyncio.to_thread(source.get_schema, table, shuffle=shuffle, **get_kwargs)
    schema = dict(schema)

    # first pop regardless to prevent
    # argument of type 'numpy.int64' is not iterable
    count = schema.pop("__len__", None)

    if not include_type:
        for field, spec in schema.items():
            if "type" in spec:
                spec.pop("type")

    empty_fields = []
    if include_min_max:
        for field, spec in schema.items():
            min_val = 0
            max_val = 0
            if "inclusiveMinimum" in spec:
                min_val = spec.pop("inclusiveMinimum")
                if isinstance(min_val, (int, float)):
                    min_val = format_float(min_val)
                if include_nans or not pd.isna(min_val):
                    spec["min"] = min_val
            if "inclusiveMaximum" in spec:
                max_val = spec.pop("inclusiveMaximum")
                if isinstance(max_val, (int, float)):
                    max_val = format_float(max_val)
                if include_nans or not pd.isna(max_val):
                    spec["max"] = max_val
            if not include_empty_fields and pd.isna(min_val) and pd.isna(max_val):
                empty_fields.append(field)
                continue
    else:
        for field, spec in schema.items():
            min_val = 0
            max_val = 0
            if "inclusiveMinimum" in spec:
                min_val = spec.pop("inclusiveMinimum")
            if "inclusiveMaximum" in spec:
                max_val = spec.pop("inclusiveMaximum")
            if "min" in spec:
                min_val = spec.pop("min")
            if "max" in spec:
                max_val = spec.pop("max")
            if not include_empty_fields and pd.isna(min_val) and pd.isna(max_val):
                empty_fields.append(field)
                continue

    num_cols = len(schema)
    for field, spec in schema.items():
        if "type" in spec:
            if not include_empty_fields and spec["type"] is None:
                spec.pop("type")
                continue
            if spec["type"] == "string":
                spec["type"] = "str"
            elif spec["type"] == "integer":
                spec["type"] = "int"
            elif spec["type"] == "number":
                spec["type"] = "num"
            elif spec["type"] == "boolean":
                spec["type"] = "bool"

        # Process enums using the extracted function
        if "enum" in spec:
            spec, is_empty = process_enums(
                spec, num_cols,
                limit=get_kwargs.get("limit"),
                include_enum=include_enum,
                reduce_enums=reduce_enums,
                include_empty_fields=include_empty_fields
            )
            if is_empty:
                empty_fields.append(field)
                continue

    # Format any other numeric values in the schema
    for field, spec in schema.items():
        for key, value in spec.items():
            if isinstance(value, (int, float)) and key not in ['type']:
                spec[key] = format_float(value)

    # remove any fields that are empty
    for field in empty_fields:
        schema[field] = "<null>"

    if count is not None and include_count:
        schema["__len__"] = count
    return schema


async def get_pipeline(**kwargs):
    """
    A wrapper be able to use asyncio.to_thread and not
    block the main thread when calling Pipeline
    """
    def get_pipeline_sync():
        return Pipeline(**kwargs)
    return await asyncio.to_thread(get_pipeline_sync)


async def get_data(pipeline):
    """
    A wrapper be able to use asyncio.to_thread and not
    block the main thread when calling pipeline.data
    """
    def get_data_sync():
        return pipeline.data
    return await asyncio.to_thread(get_data_sync)


async def describe_data(df: pd.DataFrame) -> str:
    def describe_data_sync(df):
        size = df.size
        shape = df.shape
        if size < 250:
            return df

        is_summarized = False
        if shape[0] > 5000:
            is_summarized = True
            df = df.sample(5000)

        df = df.sort_index()

        for col in df.columns:
            if isinstance(df[col].iloc[0], pd.Timestamp):
                df[col] = pd.to_datetime(df[col])

        sampled_columns = False
        if len(df.columns) > 10:
            df = df[df.columns[:10]]
            sampled_columns = True

        describe_df = df.describe(percentiles=[])
        columns_to_drop = ["min", "max"] # present if any numeric
        columns_to_drop = [col for col in columns_to_drop if col in describe_df.columns]
        df_describe_dict = describe_df.drop(columns=columns_to_drop).to_dict()

        for col in df.select_dtypes(include=["object"]).columns:
            if col not in df_describe_dict:
                df_describe_dict[col] = {}
            try:
                # Get unique values for enum processing
                unique_values = df[col].dropna().unique().tolist()
                if unique_values:
                    # Process enums just like in get_schema
                    temp_spec = {"enum": unique_values}
                    updated_spec, _ = process_enums(
                        temp_spec,
                        num_cols=len(df.columns),
                        limit=5000,  # Using the sample limit
                        include_enum=True,
                        reduce_enums=True
                    )
                    if "enum" in updated_spec:
                        df_describe_dict[col]["enum"] = updated_spec["enum"]

                df_describe_dict[col]["nunique"] = df[col].nunique()
            except Exception:
                df_describe_dict[col]["nunique"] = 'unknown'
            try:
                df_describe_dict[col]["lengths"] = {
                    "max": df[col].str.len().max(),
                    "min": df[col].str.len().min(),
                    "mean": format_float(float(df[col].str.len().mean())),
                }
            except AttributeError:
                pass

        for col in df.columns:
            if col not in df_describe_dict:
                df_describe_dict[col] = {}
            nulls = int(df[col].isnull().sum())
            if nulls > 0:
                df_describe_dict[col]["nulls"] = nulls

        # select datetime64 columns
        for col in df.select_dtypes(include=["datetime64"]).columns:
            for key in df_describe_dict[col]:
                df_describe_dict[col][key] = str(df_describe_dict[col][key])
            df[col] = df[col].astype(str)  # shorten output

        # select all numeric columns and round
        for col in df.select_dtypes(include=["int64", "float64"]).columns:
            for key in df_describe_dict[col]:
                df_describe_dict[col][key] = format_float(df_describe_dict[col][key])

        for col in df.select_dtypes(include=["float64"]).columns:
            df[col] = df[col].apply(format_float)

        return {
            "summary": {
                "total_table_cells": size,
                "total_shape": shape,
                "sampled_columns": sampled_columns,
                "is_summarized": is_summarized,
            },
            "stats": df_describe_dict,
        }

    return await asyncio.to_thread(describe_data_sync, df)


def clean_sql(sql_expr: str, dialect: str | None = None) -> str:
    """
    Cleans up a SQL expression generated by an LLM by removing
    backticks, fencing and extraneous space and semi-colons.
    """
    sql_expr = SQLRemoveSourceSeparator.apply_to(sql_expr)
    cleaned_sql = sql_expr.replace("```sql", "").replace("```", "")
    if dialect != 'bigquery':
        cleaned_sql = cleaned_sql.replace('`', '"')
    return cleaned_sql.strip().rstrip(";")


def report_error(exc: Exception, step: ChatStep, language: str = "python", context: str = ""):
    error_msg = str(exc)
    stream_details(f'\n\n```{language}\n{error_msg}\n```\n', step, title="Error")
    if context:
        step.stream(context)
    if len(error_msg) > 50:
        error_msg = error_msg[:50] + "..."
    step.failed_title = error_msg
    step.status = "failed"


def stream_details(content: Any, step: Any, title: str = "Expand for details", auto: bool = True, **stream_kwargs) -> str:
    """
    Process content to place code blocks inside collapsible details elements

    Parameters
    ----------
    content : str
        The content to format
    step : Any
        The chat step to stream the formatted content to, or NullStep if interface is None
    title : str
        The title of the details component
    auto : bool
        Whether to automatically determine what to put in details or all.
        If False, all of content will be placed in details components.

    Returns
    -------
    str
        The formatted content with code blocks in details components
    """
    if not hasattr(step, 'stream') or not hasattr(step, 'append'):
        # If the step is a NullStep without append method, just log the content
        log_debug(content)
        return content

    pattern = r'```([\w-]*)\n(.*?)```'
    last_end = 0
    if not auto:
        details = Details(content, title=title, collapsed=True, margin=(0, 20, 5, 20), sizing_mode="stretch_width")
        step.append(details, **stream_kwargs)
        return content

    for match in re.finditer(pattern, content, re.DOTALL):
        if match.start() > last_end:
            step.stream(content[last_end:match.start()], **stream_kwargs)

        language = match.group(1)
        code = match.group(2)

        details = Details(
            f"```{language}\n\n{code}\n\n```",
            title=title,
            collapsed=True,
            margin=(0, 20, 5, 20),
            sizing_mode="stretch_width",
        )
        step.append(details)
        last_end = match.end()

    if last_end < len(content):
        step.stream(content[last_end:], **stream_kwargs)
    return content


def log_debug(msg: str | list, offset: int = 24, prefix: str = "", suffix: str = "", show_sep: Literal["above", "below"] | None = None, show_length: bool = False):
    """
    Log a debug message with a separator line above and below.
    """
    terminal_cols, _ = get_terminal_size()
    terminal_cols -= offset  # Account for the timestamp and log level
    delimiter = "_" * terminal_cols
    if show_sep == "above":
        log.debug(f"\033[90m{delimiter}\033[0m")
    if prefix:
        log.debug(prefix)

    if isinstance(msg, list):
        for m in msg:
            log.debug(m)
    else:
        log.debug(msg)
    if show_length:
        log.debug(f"Characters: \033[94m{len(msg)}\033[0m")
    if suffix:
        log.debug(suffix)
    if show_sep == "below":
        log.debug(f"\033[90m{delimiter}\033[0m")


def mutate_user_message(content: str, messages: list[dict[str, str]], suffix: bool = True, wrap: bool | str = False, inplace: bool = True):
    """
    Helper to mutate the last user message in a list of messages. Suffixes the content by default, else prefixes.
    """
    mutated = False
    new = []
    for message in messages[::-1]:
        if message["role"] == "user" and not mutated:
            user_message = message["content"]
            if isinstance(wrap, bool):
                user_message = f"{user_message!r}"
            elif isinstance(wrap, str):
                user_message = f"{wrap}{user_message}{wrap}"

            if suffix:
                user_message += " " + content
            else:
                user_message = content + user_message
            mutated = True
            if inplace:
                message["content"] = user_message
                break
            else:
                message = dict(message, content=user_message)
        new.append(message)
    return messages if inplace else new[::-1]


def format_exception(exc: Exception, limit: int = 0) -> str:
    """
    Format and return a string representation of an exception.

    Parameters
    ----------
    exc : Exception
        The exception to be formatted.
    limit : int
        The maximum number of layers of traceback to include in the output.
        Default is 0.

    Returns
    -------
    str
        A formatted string describing the exception and its traceback.
    """
    e_msg = str(exc).replace('\033[1m', '<b>').replace('\033[0m', '</b>')
    tb = html.escape('\n'.join(traceback.format_exception(exc, limit=limit))).replace('\033[1m', '<b>').replace('\033[0m', '</b>')
    return f'<b>{type(exc).__name__}</b>: {e_msg}\n<pre style="overflow-y: auto">{tb}</pre>'


def cast_value(value):
    if len(value) == 1:
        value = value[0]
        if ',' in value and value.replace(',', '').isdigit():
            return [cast_value(v) for v in value.split(",")]
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            return value
    return [cast_value(v) for v in value]


def parse_huggingface_url(url: str) -> tuple[str, str, dict]:
    """Parse a HuggingFace URL to extract repo, model file, and model_kwargs.

    Args:
        url: HuggingFace URL in format https://huggingface.co/org/repo/blob/main/file.gguf?param1=value1&param2=value2

    Returns:
        tuple: (repo, model_file, model_kwargs)

    Raises:
        ValueError: If URL format is invalid
    """
    parts = url.split('/')
    hf_index = parts.index('huggingface.co')
    if len(parts) < hf_index + 4:  # Need at least org/repo after huggingface.co
        raise ValueError
    repo = f"{parts[hf_index+1]}/{parts[hf_index+2]}"
    model_file = parts[-1]  # Last component is the filename

    # Parse query parameters
    model_kwargs = {}
    if "?" in model_file:
        model_file, query_params = model_file.split('?')
        for key, value in parse_qs(query_params).items():
            model_kwargs[key] = cast_value(value)
    return repo, model_file, model_kwargs


def load_json(json_spec: str) -> dict:
    """
    Load a JSON string, handling unicode escape sequences.
    """
    return json.loads(json_spec.encode().decode('unicode_escape'))


def parse_table_slug(table: str, sources: dict[str, Source], normalize: bool = True) -> tuple[Source | None, str]:
    if SOURCE_TABLE_SEPARATOR in table:
        a_source_name, a_table = table.split(SOURCE_TABLE_SEPARATOR, maxsplit=1)
        a_source_obj = sources.get(a_source_name)
    else:
        a_source_obj = next(iter(sources.values()))
        a_table = table
    if normalize:
        a_table = a_source_obj.normalize_table(a_table)
    return a_source_obj, a_table


def truncate_string(s, max_length=20, ellipsis="..."):
    if len(s) <= max_length:
        return s
    part_length = (max_length - len(ellipsis)) // 2
    return f"{s[:part_length]}{ellipsis}{s[-part_length:]}"


def truncate_iterable(iterable, max_length=20) -> tuple[list, list, bool]:
    iterable_list = list(iterable)
    if len(iterable_list) > max_length:
        half = max_length // 2
        first_half_indices = list(range(half))
        second_half_indices = list(range(len(iterable_list) - half, len(iterable_list)))

        first_half_items = iterable_list[:half]
        second_half_items = iterable_list[-half:]

        cols_to_show = first_half_items + second_half_items
        original_indices = first_half_indices + second_half_indices
        show_ellipsis = True
    else:
        cols_to_show = iterable_list
        original_indices = list(range(len(iterable_list)))
        show_ellipsis = False
    return cols_to_show, original_indices, show_ellipsis


async def with_timeout(coro, timeout_seconds=10, default_value=None, error_message=None):
    """
    Executes a coroutine with a timeout.

    Parameters
    ----------
    coro : coroutine
        The coroutine to execute with a timeout
    timeout_seconds : float, default=10
        Maximum number of seconds to wait before timing out
    default_value : Any, default=None
        Value to return if timeout occurs
    error_message : str, default=None
        Custom error message to log on timeout

    Returns
    -------
    Any
        Result of the coroutine if it completes in time, otherwise default_value
    """
    try:
        return await asyncio.wait_for(coro, timeout=timeout_seconds)
    except asyncio.TimeoutError:
        if error_message:
            log_debug(error_message)
        return default_value
